from __future__ import annotations #  for 3.9 compatability

from typing import Protocol, TYPE_CHECKING

import numpy as np

import archnemesis as ans
import archnemesis.enums
from .datatypes.wave_range import WaveRange
from .datatypes.gas_descriptor import RadtranGasDescriptor

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)


if TYPE_CHECKING:
    N_LINES_OF_GAS = 'Number of lines for a gas isotopologue'
    N_TEMPS_OF_GAS = 'Number of temperature points for a gas isotopologue'
    

class LineDataProtocol(Protocol):
    """
    NOTE: May eventually want more abstraction either on top of this or under this so we can
    mix and match which databases provide which attributes.
    
    
    Protocol for something with the same structure as a record array with the following description:
    
        np.recarray[
            ['N_LINES_OF_GAS'],
            [
                ('NU', float), # Transition wavenumber (cm^{-1})
                ('SW', float), # transition intensity per molecule (weighted by terrestrial isotopologue abundance) (cm^{-1} molecule^{-1} cm^{-2}) at standard temperature and pressure (STP)
                ('A', float), # einstein-A coefficient for spontaneous emission (s^{-1})
                ('GAMMA_AMB', float), # ambient gas broadening coefficient (cm^{-1} atm^{-1})
                ('N_AMB', float), # temperature dependent exponent for `GAMMA_AMB` (NUMBER)
                ('DELTA_AMB', float), # ambient gas pressure induced line-shift (cm^{-1} atm^{-1})
                ('GAMMA_SELF', float), # self broadening coefficient (cm^{-1} atm^{-1})
                ('N_SELF', float), # temperature dependent exponent for `GAMMA_SELF` (NUMBER)
                ('ELOWER', float), # lower state energy (cm^{-1})
            ]
        ]
    """
    NU : np.ndarray[['N_LINES_OF_GAS'],float] # Transition wavenumber (cm^{-1})
    SW : np.ndarray[['N_LINES_OF_GAS'],float] # transition intensity (weighted by isotopologue abundance) (cm^{-1} / molec_cm^{-2})
    A : np.ndarray[['N_LINES_OF_GAS'],float] # einstein-A coeifficient (s^{-1})
    GAMMA_AMB : np.ndarray[['N_LINES_OF_GAS'],float] # ambient gas broadening coefficient (cm^{-1} atm^{-1})
    N_AMB : np.ndarray[['N_LINES_OF_GAS'],float] # temperature dependent exponent for `GAMMA_AMB` (NUMBER)
    DELTA_AMB : np.ndarray[['N_LINES_OF_GAS'],float] # ambient gas pressure induced line-shift (cm^{-1} atm^{-1})
    GAMMA_SELF : np.ndarray[['N_LINES_OF_GAS'],float] # self broadening coefficient (cm^{-1} atm^{-1})
    N_SELF : np.ndarray[['N_LINES_OF_GAS'],float] # temperature dependent exponent for `GAMMA_SELF` (NUMBER)
    ELOWER : np.ndarray[['N_LINES_OF_GAS'],float] # lower state energy (cm^{-1})


class PartitionFunctionDataProtocol(Protocol):
    """
    NOTE: May eventually want more abstraction either on top of this or under this so we can
    mix and match databases providing partition functions for different ranges of temperatures.
    
    Protocol for something with the same structure as a record array with the following description:
    
        np.recarray[
            ['N_TEMPS_OF_GAS'],
            [
                ('TEMP', float), # Temperature of tablulated partition function (Kelvin)
                ('Q', float), # Tabulated partition function value
            ]
        ]
    """
    TEMP : np.ndarray[['N_TEMPS_OF_GAS'],float] # Temperature of tablulated partition function (Kelvin)
    Q : np.ndarray[['N_TEMPS_OF_GAS'],float] # Tabulated partition function value


class LineDatabaseProtocol(Protocol):
    def __repr__(self):
        """
        Returns a string that represents the current state of the class
        """
        return f'{self.__class__.__name__}(instance_id={id(self)}, local_storage_dir={self.local_storage_dir})'
    
    @property
    def ready(self) -> bool:
        """
        Returns True if the database is ready to use, False otherwise.
        """
        raise NotImplementedError
    
    @property
    def local_storage_dir(self) -> str:
        """
        Returns the directory the local database is stored in
        """
        raise NotImplementedError
    
    @local_storage_dir.setter
    def local_storage_dir(self, value : str) -> None:
        """
        Sets the directory the local database is stored in
        """
        raise NotImplementedError
    
    def purge(self):
        """
        Remove all cached data from this database and make it so the database must be reinitalised
        """
        raise NotImplementedError
    
    def get_line_data(
            self, 
            gas_descriptors : tuple[RadtranGasDescriptor,...], 
            wave_range : WaveRange, 
            ambient_gas : ans.enums.AmbientGas
        ) -> dict[RadtranGasDescriptor, None | LineDataProtocol]:
        """
        Returns a dictionary where the keys are the `gas_descriptors`, and the values
        are the line data for the `gas_descriptor` in the specified `wave_range` with
        the specified `ambient_gas`.
        
        ## ARGUMENTS ##
        
            gas_descriptors : tuple[RadtranGasDescriptor,...]
                A tuple of `RadtranGasDescriptor` instances. RadtranGasDescriptor is a class wrapper around
                a pair of (gas_id, iso_id) values that use radtran id numbers. Determines which gasses
                the data is retrieved for.
                
            wave_range : WaveRange
                The range of wavelengths/wavenumbers to retrieve data for.
                
            ambient_gas : ans.enums.AmbientGas
                The ambient gas to use when retrieving data.
        
        ## RETURNS  ##
            
            line_data : dict[RadtranGasDescriptor, None | LineDataProtocol]
                A dictionary where the keys are the passed `gas_descriptors` and the values are objects that follow
                the `LineDataProtocol`. If gas is not present in database, associates to None. The
                `LineDataProtocol` is defined as any object that has the following attributes:
                
                    NU : np.ndarray[['N_LINES_OF_GAS'],float]
                        Transition wavenumber (cm^{-1})
                    
                    SW : np.ndarray[['N_LINES_OF_GAS'],float]
                        Transition intensity (weighted by isotopologue abundance) (cm^{-1} / molec_cm^{-2})
                    
                    A : np.ndarray[['N_LINES_OF_GAS'],float]
                        Einstein-A coeifficient (s^{-1})
                    
                    GAMMA_AMB : np.ndarray[['N_LINES_OF_GAS'],float]
                        Ambient gas broadening coefficient (cm^{-1} atm^{-1})
                    
                    N_AMB : np.ndarray[['N_LINES_OF_GAS'],float]
                        Temperature dependent exponent for `GAMMA_AMB` (NUMBER)
                    
                    DELTA_AMB : np.ndarray[['N_LINES_OF_GAS'],float]
                        Ambient gas pressure induced line-shift (cm^{-1} atm^{-1})
                    
                    GAMMA_SELF : np.ndarray[['N_LINES_OF_GAS'],float]
                        Self broadening coefficient (cm^{-1} atm^{-1})
                    
                    ELOWER : np.ndarray[['N_LINES_OF_GAS'],float]
                        Lower state energy (cm^{-1})
            
        """
        raise NotImplementedError



class PartitionFunctionDatabaseProtocol(Protocol):
    def __repr__(self):
        """
        Returns a string that represents the current state of the class
        """
        return f'{self.__class__.__name__}(instance_id={id(self)}, local_storage_dir={self.local_storage_dir})'
    
    @property
    def ready(self) -> bool:
        """
        Returns True if the database is ready to use, False otherwise.
        """
        raise NotImplementedError
    
    @property
    def local_storage_dir(self) -> str:
        """
        Returns the directory the local database is stored in
        """
        raise NotImplementedError
    
    @local_storage_dir.setter
    def local_storage_dir(self, value : str) -> None:
        """
        Sets the directory the local database is stored in
        """
        raise NotImplementedError
    
    def purge(self):
        """
        Remove all cached data from this database and make it so the database must be reinitalised
        """
        raise NotImplementedError
    
    
    def get_partition_function_data(
            self, 
            gas_descriptors : tuple[RadtranGasDescriptor,...]
        ) -> dict[RadtranGasDescriptor, None | PartitionFunctionDataProtocol]:
        """
        Returns a dictionary where the keys are the `gas_descriptors`, and the values
        are the partition function data for the `gas_descriptor`.
        
        ## ARGUMENTS ##
        
            gas_descriptors : tuple[RadtranGasDescriptor,...]
                A tuple of `RadtranGasDescriptor` instances. RadtranGasDescriptor is a class wrapper around
                a pair of (gas_id, iso_id) values that use radtran id numbers. Determines which gasses
                the data is retrieved for.
        
        ## RETURNS  ##
            
            partition_function_data : dict[RadtranGasDescriptor, None | PartitionFunctionDataProtocol]
                A dictionary where the keys are the passed `gas_descriptors` and the values are objects that follow
                the `PartitionFunctionDataProtocol`. If the gas is not present in the database the associated value
                is None. The `PartitionFunctionDataProtocol` is defined as any object that has the following attributes:
                
                    TEMP : np.ndarray[['N_TEMPS_OF_GAS'],float]
                        Temperature of tablulated partition function (Kelvin)
                    
                    Q : np.ndarray[['N_TEMPS_OF_GAS'],float]
                        Tabulated partition function value
            
        """
        raise NotImplementedError



from __future__ import annotations #  for 3.9 compatability

import numpy as np
from scipy.special import (
    voigt_profile, 
    #wofz, 
    #dawsn
)

SQRT_2 = np.sqrt(2)
SQRT_PI = np.sqrt(np.pi)
SQRT_2PI = SQRT_2 * SQRT_PI
SQRT_2log2 = np.sqrt(2*np.log(2))
SQRT_log2 = np.sqrt(np.log(2))



def voigt_ch4_H2_ambient(
        delta_wn : np.ndarray, 
        alpha_d : float, 
        gamma_l : float
    ) -> np.ndarray:
    
    
    # NOTE: Unsure where the log(2) factor comes from but do not get the same answer as existing LBL tables without it.
    #       it could also be a factor of 1/sqrt(2), they are very similar numbers.
    #return voigt_profile(delta_wn, alpha_d, np.log(2)*gamma_l) 

    # NOTE: Unsure where the 1/sqrt(2) factor comes from but do not get the same answer as existing LBL tables without it.
    #       it could also be a factor of log(2), they are very similar numbers.
    return voigt_profile(delta_wn, alpha_d/SQRT_2, gamma_l/SQRT_2)


def voigt(
        delta_wn : np.ndarray, 
        alpha_d : float, 
        gamma_l : float
    ) -> np.ndarray:
    """
    Compute Voigt profile using Humlicek's algorithm
    
    ## ARGUMENTS ##
        delta_wn : np.ndarray
            Wavenumber difference from line center
        alpha_d : float
            Doppler width (gaussian HWHM)
        gamma_l : float
            Lorentz width (cauchy-lorentz HWHM)
    
    ## CALC ##
    
        sigma : gassian standard deviation
        gamma : lorentz half-width-half-maximum
        
        Gaussian
        --------
        G(dv, sigma) = 1/(sqrt(2*PI) * sigma) EXP(-dv^2 / (2*sigma)^2)
        
        Lorentz
        -------
        L(dv, gamma) = gamma / (PI * (gamma^2 + dv^2))
        
        Voigt
        -----
        V(dv, sigma, gamma) = Re[WOFZ(z)] / (sqrt(2*pi) * sigma)
        
        where:
            z = (dv + gamma * i)/(sqrt(2) * sigma)
        
        
        
        alpha_d : gaussian HWHM = sqrt(2 ln(2)) * sigma
        --->  sigma = aD / sqrt(2 ln(2))
        
        gamma_l : cauchy-lorents HWHM = gamma
    """
    
    sigma = alpha_d / SQRT_2log2
    return voigt_profile(delta_wn, sigma, gamma_l)


def hartmann_empirical_infrared_ch4_h2_broadening(
        delta_wn : np.ndarray, 
        alpha_d : float, 
        gamma_l : float
    ) -> np.ndarray:
    """
    Compute CH4 Voigt lineshape modified for sub-lorentzian line
    wings. Original coefficients recommended by Hartmann (2002)
    """
    chi = np.ones_like(delta_wn)
    abs_delta_wn = np.abs(delta_wn)
    mask_26 = abs_delta_wn < 26
    mask_60 = abs_delta_wn < 60
    
    mask_a = ~mask_26 & mask_60
    
    chi[mask_a] = 8.72*np.exp(-abs_delta_wn[mask_a]/12.0)
    chi[~mask_60] = 0.0684*np.exp(-abs_delta_wn[~mask_60]/ 393.0)
    
    
    return chi*voigt(delta_wn, alpha_d, gamma_l)


import pytest  
import archnemesis as ans
import numpy as np
import os
curr = os.getcwd()


def test_makephase_imie0():  
    '''
    Test MakePhase program to calculate the extinction coefficient, single scattering albedo and phase function using the Mie theory
    
    We make this test following the refractive index of Mars dust and assuming the following particle size distributions:
    - Log-normal distribution with a = 1.0, b = 0.1
    - Standard gamma distribution with a = 2.0, b = 0.1
    - Single particle size with r = 1.5
    - MCS Modified standard gamma distribution with a = 2.0, b = 0.1, c = 6.
    
    
    In this case, IMIE = 0, which means that the phase function is then fit using a double Henyey-Greenstein function.
    '''
    
    # CALCULATING THE EXTINCTION COEFFICIENT, SINGLE SCATTERING ALBEDO AND PHASE FUNCTION USING THE MIE THEORY WITH ARCHNEMESIS
    ##################################################################################################################################
    
    
    #Declaring class
    Scatter = ans.Scatter_0()
    
    #Defining the characteristics of the class
    Scatter.ISPACE = 1   #Units of the calculations (0 - Wavenumber in cm-1 ; 1 - Wavelength in um)
    Scatter.IMIE = 0     #Double Henyey-Greenstein phase function fit

    NDUST = 4      #Number of aerosol populations that we want to include in our atmosphere
    wavel = np.arange(0.5,2.5+0.5,0.5)
    theta = np.array([0.,1.,2.,3.,4.,5.,7.5,10.,12.5,15.,17.5,20.,25.,30.,35.,40.,50.,60.,70.,80.,90.,100.,110.,120.,130.,140.,145.,150.,155.,160.,162.5,165.,167.5,170.,172.5,175.,176.,177.,178.,179.,180.])
    NWAVE = len(wavel)    #Number of spectral points
    NTHETA = len(theta)
    
    #Now we initialise the arrays that will be filled with the calculations
    Scatter.initialise_arrays(NDUST,NWAVE,NTHETA)
    Scatter.WAVE = wavel
    Scatter.THETA = theta
    
    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs for our standard gamma particle size distribution
    iscat = 2  #Log-normal distribution
    pars = np.array([1.0,0.1])

    idust = 0    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)

    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs for our standard gamma particle size distribution
    iscat = 1  #Standard gamma distribution
    a = 2.0 ; b = 0.1 ; alpha = (1-3.*b)/b
    pars = np.array([a,b,alpha])

    idust = 1    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)     



    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs 
    iscat = 4  #Single particle size
    r0 = 1.5
    pars = np.array([r0])

    idust = 2    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)



    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs for our standard gamma particle size distribution
    iscat = 3  #Modified gamma distribution
    a = 2. ; b = 0.1 ; c = 6.
    pars = np.array([a,b,c])

    idust = 3    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)
    
    Scatter.check_phase_norm()
 
    # READING THE RESULTS OF THE CALCULATIONS FROM NEMESIS
    ##################################################################################################################################
    
    ScatterN = ans.Scatter_0()
    ScatterN.IMIE = 0
    test_dir = os.path.join(ans.archnemesis_path(), 'tests/files/Makephase/')
    os.chdir(test_dir) #Changing directory to read files
    ScatterN.read_xsc('nemesis_mars_dust')
    ScatterN.read_hgphase()
    
    assert np.allclose(Scatter.KEXT, ScatterN.KEXT, rtol=1.0e-3)
    

  
def test_makephase_imie1():  
    '''
    Test MakePhase program to calculate the extinction coefficient, single scattering albedo and phase function using the Mie theory
    
    We make this test following the refractive index of Mars dust and assuming the following particle size distributions:
    - Log-normal distribution with a = 1.0, b = 0.1
    - Standard gamma distribution with a = 2.0, b = 0.1
    - Single particle size with r = 1.5
    - MCS Modified standard gamma distribution with a = 2.0, b = 0.1, c = 6.
    '''
    
    # CALCULATING THE EXTINCTION COEFFICIENT, SINGLE SCATTERING ALBEDO AND PHASE FUNCTION USING THE MIE THEORY WITH ARCHNEMESIS
    ##################################################################################################################################
    
    
    #Declaring class
    Scatter = ans.Scatter_0()
    
    #Defining the characteristics of the class
    Scatter.ISPACE = 1   #Units of the calculations (0 - Wavenumber in cm-1 ; 1 - Wavelength in um)
    Scatter.IMIE = 1     #Phase function defined explicitly

    NDUST = 4      #Number of aerosol populations that we want to include in our atmosphere
    wavel = np.arange(0.5,2.5+0.5,0.5)
    theta = np.array([0.,1.,2.,3.,4.,5.,7.5,10.,12.5,15.,17.5,20.,25.,30.,35.,40.,50.,60.,70.,80.,90.,100.,110.,120.,130.,140.,145.,150.,155.,160.,162.5,165.,167.5,170.,172.5,175.,176.,177.,178.,179.,180.])
    NWAVE = len(wavel)    #Number of spectral points
    NTHETA = len(theta)
    
    #Now we initialise the arrays that will be filled with the calculations
    Scatter.initialise_arrays(NDUST,NWAVE,NTHETA)
    Scatter.WAVE = wavel
    Scatter.THETA = theta
    
    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs for our standard gamma particle size distribution
    iscat = 2  #Log-normal distribution
    pars = np.array([1.0,0.1])

    idust = 0    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)

    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs for our standard gamma particle size distribution
    iscat = 1  #Standard gamma distribution
    a = 2.0 ; b = 0.1 ; alpha = (1-3.*b)/b
    pars = np.array([a,b,alpha])

    idust = 1    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)     



    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs 
    iscat = 4  #Single particle size
    r0 = 1.5
    pars = np.array([r0])

    idust = 2    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)



    #Defining the inputs for our calculations of this particle aerosol distribution
    ################################################################################################

    #Reading the refractive index from the dictionary (Mars dust)
    Scatter.read_refind(1)

    #Defining the inputs for our standard gamma particle size distribution
    iscat = 3  #Modified gamma distribution
    a = 2. ; b = 0.1 ; c = 6.
    pars = np.array([a,b,c])

    idust = 3    #The index of the aerosol populations in the class that this calculation corresponds to (from 0 to NDUST-1)
    Scatter.makephase(idust,iscat,pars)
    
    Scatter.check_phase_norm()
 
    # READING THE RESULTS OF THE CALCULATIONS FROM NEMESIS
    ##################################################################################################################################
    
    ScatterN = ans.Scatter_0()
    ScatterN.IMIE = 1
    test_dir = os.path.join(ans.archnemesis_path(), 'tests/files/Makephase/')
    os.chdir(test_dir) #Changing directory to read files
    ScatterN.read_xsc('nemesis_mars_dust')
    ScatterN.read_phase()
    
    assert np.allclose(Scatter.KEXT, ScatterN.KEXT, rtol=1.0e-3)
    


from __future__ import annotations #  for 3.9 compatability

"""
The HITRAN API "HAPI" has quite a few quirks that mean it does
not behave as expected much of the time. Therefore I've created
this wrapper to handle its import.

Ideally you would only ever use this to import HAPI, and never do it
directly.

USAGE:
    `import archnemesis.database.wrappers.hapi as hapi`
"""
import sys
import builtins
from io import StringIO
import inspect


import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.INFO)

__HAPI_string_buffer = StringIO()

if "hapi" in sys.modules:
    _lgr.warning('\n'.join(('',
        'Module "hapi" that provides the HITRAN API should only be imported by the ',
        '`archnemesis.database.wrappers.hapi` module.',
        '',
        'Use `import archnemesis.database.wrappers.hapi as hapi` to import HAPI in ',
        'any other module.',
        '',
        'This is because HAPI works quite differently from how python modules are ',
        'expected to work. And quite differently from how database managers work ',
        'in particular. Using the normal import method *should* not break anything ',
        'but may result in lots of extraenous output to stdout.',
    )))

# Need to redefine print when we import so we don't get loads of
# superflous output that we cannot get rid of.
__builtin_print = builtins.print

def __HAPI_print_intercept(*args, **kwargs) -> None:
    if kwargs.get('file', sys.stdout) is sys.stdout:
        if _lgr.level <= logging.DEBUG:
            kwargs['file'] = __HAPI_string_buffer
            __builtin_print(*args, **kwargs)
            __HAPI_string_buffer.truncate()
            s = __HAPI_string_buffer.getvalue()
            _lgr.debug(s[:-1] if s.endswith('\n') else s, stacklevel=2)
            __HAPI_string_buffer.seek(0)
    else:
        __builtin_print(*args, **kwargs)

builtins.print = __HAPI_print_intercept


HAPI_MODULE_INSTALLED = True
try:
    # Must import this way to avoid hapi.__init__ shadowing various bits of the module.
    import hapi.hapi as hapi
except ModuleNotFoundError:
    _lgr.warning('HAPI (hitran-api) module was not installed, therefore HITRAN online backend is NOT available. All attempts to use the backend will raise a `ModuleNotFoundError` exception')
    HAPI_MODULE_INSTALLED = False
finally:
    builtins.print = __builtin_print

if not HAPI_MODULE_INSTALLED:
    def __getattr__(name):
        raise ModuleNotFoundError('HAPI (hitran-api) module was not installed, therefore HITRAN online backend is NOT available.')
    
    def __setattr__(name, value):
        raise ModuleNotFoundError('HAPI (hitran-api) module was not installed, therefore HITRAN online backend is NOT available.')

else:
    if _lgr.level <= logging.DEBUG:
        hapi.VARIABLES['DISPLAY_FETCH_URL'] = True
    
    
    
    hapi.LOCAL_TABLE_CACHE = dict()

    from . import hapi_monkey_patches as HMP

    for name, monkey_patch_fn in inspect.getmembers(HMP, lambda x: inspect.isfunction(x) and x.__name__.endswith('_MONKEYPATCH')):
        hapi_name = name[:-len('_MONKEYPATCH')]
        _lgr.debug(f'Monkey patching function `hapi.{hapi_name}(...)`')
        setattr(hapi, hapi_name, monkey_patch_fn)

    # Use these two functions to pass everything through to the "real" `hapi` module.
    def __getattr__(name):
        wrapped = None
        
        wrapped = getattr(hapi, name)
        
        
        if not callable(wrapped) or name=='describeTable':
            return wrapped
        
        def wrapper(*args, **kwargs):
            builtins.print = __HAPI_print_intercept
            try:
                result = wrapped(*args, **kwargs)
            finally:
                builtins.print = __builtin_print
            return result
        return wrapper

    def __setattr__(name, value):
        setattr(hapi, name, value)



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Layer_0.py - Object to represent the layers of the atmosphere.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability
import numpy as np
import matplotlib.pyplot as plt
from numba import jit

from archnemesis.helpers import h5py_helper
from archnemesis.enums import (
    LayerType,
    LayerIntegrationScheme,
    InterpolationMethod,
)

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

AVOGAD = 6.02214076e23
"""
Object to store layering scheme settings and averaged properties of each layer.
"""
class Layer_0:
    def __init__(self, RADIUS=None, LAYHT=0, LAYTYP=LayerType.EQUAL_LOG_PRESSURE, LAYINT=LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE, NLAY=20, NINT=101,
                LAYANG=0.0, H_base=None, P_base=None):
        """
        After creating a Layer object, call the method
            integrate(self, H, P, T, LAYANG, ID, VMR) to calculate
            averaged layer properties.

        Inputs
        ------

        @param RADIUS: real
            Reference planetary radius where H=0.  Usually at surface for
            terrestrial planets, or at 1 bar pressure level for gas giants.
        @param LAYHT: real
            Height of the base of the lowest layer. Default 0.0.
        @param LAYTYP: LayerType (enum)
            Integer specifying how to split up the layers. Default LayerType.EQUAL_LOG_PRESSURE.
            LayerType.EQUAL_PRESSURE = by equal changes in pressure
            LayerType.EQUAL_LOG_PRESSURE = by equal changes in log pressure
            LayerType.EQUAL_HEIGHT = by equal changes in height
            LayerType.EQUAL_PATH_LENGTH = by equal changes in path length at LAYANG
            LayerType.BASE_PRESSURE = layer base pressure levels specified by P_base
            LayerType.BASE_HEIGHT = layer base height levels specified by H_base
            Note BASE_PRESSURE and BASE_HEIGHT force NLAY = len(P_base) or len(H_base).
        @param LAYINT: LayerIntegrationScheme (enum)
            Layer integration scheme
            LayerIntegrationScheme.MID_PATH = use properties at mid path
            LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE = use absorber amount weighted average values
        @param LAYANG: real
            Angle from zenith at which to split the layers (Default is 0.0)
        @param NLAY: int
            Number of layers to split the atmosphere into. Default 20.
        @param NINT: int
            Number of integration points to be used if LAYINT=LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE.
        @param H_base: 1D array
            Heights of the layer bases defined by user. Default None.
        @param P_base: 1D array
            Pressures of the layer bases defined by user. Default None.


        Attributes
        ------
    
        @param BASEH: 1D array (NLAY)
            Base altitude of each of the layers in the atmosphere (m)
        @param BASEP: 1D array (NLAY)
            Base pressure of each of the layers in the atmosphere (Pa)
        @param DELH: 1D array (NLAY)
            Width of each of the layers in the atmosphere (m)
        @param HEIGHT: 1D array (NLAY)
            Effective altitude each of the layers in the atmosphere (m)
        @param PRESS: 1D array (NLAY)
            Effective pressure of each layer (Pa) (the one at which the radiative transfer calculations are performed)
        @param TEMP: 1D array (NLAY)
            Effective temperature of each layer (Pa) (the one at which the radiative transfer calculations are performed)
        @param TOTAM: 1D array (NLAY)
            Gaseous line-of-sight column density in each layer (particles/m2)
        @param AMOUNT: 2D array (NLAY,NVMR)
            Gaseous line-of-sight column density of each gas in each layer (particles/m2)
        @param PP: 2D array (NLAY,NVMR)
            Partial pressure of each gas in each layer (Pa)
        @param CONT: 2D array (NLAY,NDUST)
            Gaseous line-of-sight column density of each dust population in each layer (particles/m2)

        @param DTE: 2D array (NLAY,NP)
            Matrix relating the temperatures in each layer wrt those in the input profiles
        @param DTE: 2D array (NLAY,NP)
            Matrix relating the gaseous abundances in each layer wrt those in the input profiles
        @param DCO: 2D array (NLAY,NP)
            Matrix relating the dust abundances in each layer wrt those in the input profiles
        @param DPH: 2D array (NLAY,NP)
            Matrix relating the para-H2 fractions in each layer wrt those in the input profiles

        @param TAURAY: 2D array (NWAVE,NLAY)
            Rayleigh scattering optical depth
        @param TAUSCAT: 2D array (NWAVE,NLAY)
            Aerosol scattering optical depth
        @param TAUCLSCAT: 3D array (NWAVE,NLAY,NDUST)
            Aerosol scattering optical depth by each aerosol type
        @param TAUDUST: 2D array (NWAVE,NLAY)
            Aerosol extinction optical depth (absorption + scattering)
        @param TAUCIA: 2D array (NWAVE,NLAY)
            CIA optical depth
        @param TAUGAS: 3D array (NWAVE,NG,NLAY)
            Gas optical depth
        @param TAUTOT: 3D array (NWAVE,NG,NLAY)
            Total optical depth


        Methods
        -------

        Layer_0.assess()
        Layer_0.write_hdf5()
        Layer_0.read_hdf5()
        Layer_0.integrate()
        Layer_0.integrateg()
        Layer_0.layer_split()
        Layer_0.layer_average()
        Layer_0.layer_averageg()

        """
        self.RADIUS = RADIUS
        #self.LAYTYP = LAYTYP
        self.NLAY = NLAY
        #self.LAYINT = LAYINT
        self.NINT = NINT
        self.LAYHT = LAYHT
        self.H_base = H_base
        self.P_base = P_base
        self.LAYANG = None     #Angle used to split the layers. Usually it is zero unless it is a limb calculation that it is 90.
        
        #Properties of the layers
        self.BASEH = None      #(NLAY) Base altitude of each layer (m)
        self.BASEP = None      #(NLAY) Base pressure of each layer (Pa)
        self.DELH = None       #(NLAY) Width of each layer (m)
        self.HEIGHT = None     #(NLAY) Effective height of each layer (m) 
        self.PRESS = None      #(NLAY) Effective pressure of each layer (Pa) (the one at which the radiative transfer calculations are performed)
        self.TEMP = None       #(NLAY) Effective temperature of each layer (K) (the one at which the radiative transfer calculations are performed)
        self.TOTAM = None      #(NLAY) Gaseous line-of-sight column density in each layer (m-2)
        self.AMOUNT = None     #(NLAY,NVMR) Gaseous line-of-sight column density of each gas in each layer (m-2)
        self.PP = None         #(NLAY,NVMR) Effective partial pressure of each gas in each layer (Pa)
        self.CONT = None       #(NLAY,NDUST)  Line-of-sight column density of each dust population in each layer (particles/m2)
        self.FRAC = None     #(NLAY) Fraction of para-h2 in each layer
        
        #Matrices relating the properties at the reference input profiles wrt the properties of each layer 
        self.DTE = None        #(NLAY,NP)
        self.DAM = None        #(NLAY,NP)
        self.DCO = None        #(NLAY,NP)
        self.DPH = None        #(NLAY,NP)

        #Optical depths in each layer
        self.NWAVE = None  #Number of calculation wavelengths
        self.TAURAY = None  #(NWAVE,NLAY) Rayleigh scattering optical depth
        self.TAUSCAT = None #(NWAVE,NLAY) Aerosol scattering optical depth
        self.TAUDUST = None #(NWAVE,NLAY) Aerosol absorption + scattering optical depth
        self.TAUCIA = None  #(NWAVE,NLAY) CIA optical depth
        self.TAUGAS = None  #(NWAVE,NG,NLAY) Gas optical depth
        self.TAUTOT = None  #(NWAVE,NG,NLAY) Total optical depth

        #Units of the dust
        self.DUST_UNITS_FLAG = None   #If None, dust units are expected to be particles m-3. If -1 they are expected to be particles per gram of atmosphere

        # private attributes
        self._laytyp = None
        self._layint = None
        
        # set properties
        self.LAYTYP = LAYTYP
        self.LAYINT = LAYINT
    
    @property
    def LAYTYP(self) -> LayerType:
        return self._laytyp
    
    @LAYTYP.setter
    def LAYTYP(self, value):
        self._laytyp = LayerType(value)
    
    @property
    def LAYINT(self) -> LayerIntegrationScheme:
        return self._layint
    
    @LAYINT.setter
    def LAYINT(self, value):
        self._layint = LayerIntegrationScheme(value)

    ####################################################################################################

    def assess(self):
        """
        Assess whether the different variables have the correct dimensions and types
        """

        #Checking some common parameters to all cases
        assert np.issubdtype(type(self.NLAY), np.integer) == True , \
            'NLAY must be int'
        assert self.NLAY > 0 , \
            'NLAY must be >0'
        
        #Checking some common parameters to all cases
        assert isinstance(self.LAYTYP, (int, LayerType)), \
            'LAYTYP must be int or LayerType'
        assert self.LAYTYP >= LayerType.EQUAL_PRESSURE , \
            'LAYTYP must be >= LayerType.EQUAL_PRESSURE'
        assert self.LAYTYP <= LayerType.BASE_HEIGHT , \
            'LAYTYP must be <= LayerType.BASE_HEIGHT'
        
        #Checking some common parameters to all cases
        assert isinstance(self.LAYINT, (int, LayerIntegrationScheme)), \
            'LAYINT must be int or LayerIntegrationScheme'
        assert self.LAYINT >= LayerIntegrationScheme.MID_PATH , \
            'LAYINT must be >= LayerIntegrationScheme.MID_PATH'
        assert self.LAYINT <= LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE , \
            'LAYINT must be <= LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE'
        
        if self.LAYTYP==LayerType.BASE_PRESSURE:
            assert len(self.P_base) == self.NLAY , \
                'P_base must have size (NLAY) if LAYTYP=LayerType.BASE_PRESSURE'
        
        if self.LAYTYP==LayerType.BASE_HEIGHT:
            assert len(self.H_base) == self.NLAY , \
                'H_base must have size (NLAY) if LAYTYP=LayerType.BASE_HEIGHT'


    ####################################################################################################
    
    def summary_info(self):
        """
        Print summary information about the class in the screen
        """
        
        _lgr.info(f'Number of layers ::  {(self.NLAY)}')

        if self.LAYTYP==LayerType.EQUAL_PRESSURE:
            _lgr.info('Layers calculated by equal changes in pressure')
        elif self.LAYTYP==LayerType.EQUAL_LOG_PRESSURE:
            _lgr.info('Layers calculated by equal changes in log pressure')
        elif self.LAYTYP==LayerType.EQUAL_HEIGHT:
            _lgr.info('Layers calculated by equal changes in altitude')
        elif self.LAYTYP==LayerType.EQUAL_PATH_LENGTH:
            _lgr.info('Layers calculated by equal changes in line-of-sight path intervals')
        elif self.LAYTYP==LayerType.BASE_PRESSURE:
            _lgr.info('Layers calculated by specified base pressures')
        elif self.LAYTYP==LayerType.BASE_HEIGHT:
            _lgr.info('Layers calculated by specified base altitudes')

        if self.LAYINT==LayerIntegrationScheme.MID_PATH:
            _lgr.info('Layer properties calculated at the middle of the layer')
        elif self.LAYINT==LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE:
            _lgr.info('Layer properties calculated through mass weighted averages')
            
        if self.BASEH is not None:
            _lgr.info(f"BASEH(km) {('BASEP(bar)','DELH(km)','P(bar)','T(K)','TOTAM(m-2)','DUST_TOTAM(m-2)')}")
            for i in range(self.NLAY):
                _lgr.info(self.BASEH[i]/1.0e3,self.BASEP[i]/1.0e5,self.DELH[i]/1.0e3,self.PRESS[i]/1.0e5,self.TEMP[i],np.sum(self.AMOUNT[i,:]),np.sum(self.CONT[i,:]))


    ####################################################################################################

    def write_hdf5(self,runname):
        """
        Write the information about the layering of the atmosphere in the HDF5 file
        """

        import h5py

        self.assess()

        #Writing the information into the HDF5 file
        with h5py.File(runname+'.h5', 'a') as f:
            #Checking if Layer already exists
            if ('/Layer' in f)==True:
                del f['Layer']   #Deleting the Layer information that was previously written in the file

            grp = f.create_group("Layer")

            #Writing the layering type
            dset = h5py_helper.store_data(grp, 'LAYTYP', int(self.LAYTYP))
            dset.attrs['title'] = "Layer splitting calculation type"
            if self.LAYTYP==LayerType.EQUAL_PRESSURE:
                dset.attrs['type'] = 'Split layers by equal changes in pressure'
            elif self.LAYTYP==LayerType.EQUAL_LOG_PRESSURE:
                dset.attrs['type'] = 'Split layers by equal changes in log-pressure'
            elif self.LAYTYP==LayerType.EQUAL_HEIGHT:
                dset.attrs['type'] = 'Split layers by equal changes in height'
            elif self.LAYTYP==LayerType.EQUAL_PATH_LENGTH: 
                dset.attrs['type'] = 'Split layers by equal changes in path length (at LAYANG)'
            elif self.LAYTYP==LayerType.BASE_PRESSURE:
                dset.attrs['type'] = 'Split layers by defining base pressure of each layer'
            elif self.LAYTYP==LayerType.BASE_HEIGHT:
                dset.attrs['type'] = 'Split layers by defining base altitude of each layer'

            #Writing the layering integration type
            dset = h5py_helper.store_data(grp, 'LAYINT', int(self.LAYINT))
            dset.attrs['title'] = "Layer integration calculation type"
            if self.LAYINT==LayerIntegrationScheme.MID_PATH:
                dset.attrs['type'] = 'Layer properties calculated at mid-point between layer boundaries'
            if self.LAYINT==LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE:
                dset.attrs['type'] = 'Layer properties calculated by weighting in terms of atmospheric mass (i.e. Curtis-Godson path)'
        
            #Writing the number of layers
            dset = h5py_helper.store_data(grp, 'NLAY', self.NLAY)
            dset.attrs['title'] = "Number of layers"

            #Writing the altitude of bottom layer
            dset = h5py_helper.store_data(grp, 'LAYHT', self.LAYHT)
            dset.attrs['title'] = "Altitude at base of bottom layer"
            dset.attrs['units'] = "m"

            #Writing the base properties of the layers in special cases
            if self.LAYTYP==LayerType.BASE_PRESSURE:
                dset = h5py_helper.store_data(grp, 'P_base', self.P_base)
                dset.attrs['title'] = "User defined pressure at the base of each layer"
                dset.attrs['units'] = "Pressure / Pa"
            
            if self.LAYTYP==LayerType.BASE_HEIGHT:
                dset = h5py_helper.store_data(grp, 'H_base', self.H_base)
                dset.attrs['title'] = "User defined altitude at the base of each layer"
                dset.attrs['units'] = "Altitude / m"
            
            dset = h5py_helper.store_data(grp, 'BASEH', self.BASEH)
            dset.attrs['title'] = "Altitude at the base of each layer"
            dset.attrs['units'] = "Altitude / m"
            
            dset = h5py_helper.store_data(grp, 'BASEP', self.BASEP)
            dset.attrs['title'] = "Altitude at the base of each layer"
            dset.attrs['units'] = "Pressure / Pa"
            
            dset = h5py_helper.store_data(grp, 'TAUTOT', self.TAUTOT)
            dset.attrs['title'] = 'Total opacity of each wavelength at each G-ordinate at each layer. NOTE: G-ordinate is an interval of cumulative fraction of absorption coefficient (see correlated-k approximation)'
            dset.attrs['units'] = 'cm^{-2}'
        
        return
    
    
    def read_hdf5(self,runname):
        """
        Read the Layer properties from an HDF5 file
        """

        import h5py

        with h5py.File(runname+'.h5', 'r') as f:
            #Checking if Surface exists
            e = "/Layer" in f
            if e==False:
                raise ValueError('error :: Layer is not defined in HDF5 file')
            else:

                self.NLAY = h5py_helper.retrieve_data(f, 'Layer/NLAY', np.int32, default=20)
                self.LAYTYP = h5py_helper.retrieve_data(f, 'Layer/LAYTYP', lambda x:  LayerType(np.int32(x)), default=LayerType.EQUAL_LOG_PRESSURE)
                self.LAYINT = h5py_helper.retrieve_data(f, 'Layer/LAYINT', lambda x:  LayerIntegrationScheme(np.int32(x)), default=LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE)
                self.LAYHT = h5py_helper.retrieve_data(f, 'Layer/LAYHT', np.float64)
                if self.LAYTYP==LayerType.BASE_PRESSURE:
                    self.P_base = h5py_helper.retrieve_data(f, 'Layer/P_base', np.array)
                if self.LAYTYP==LayerType.BASE_HEIGHT:
                    self.H_base = h5py_helper.retrieve_data(f, 'Layer/H_base', np.array)
                self.BASEH = h5py_helper.retrieve_data(f, 'Layer/BASEH', np.array)
                self.BASEP = h5py_helper.retrieve_data(f, 'Layer/BASEP', np.array)
                self.TAUTOT = h5py_helper.retrieve_data(f, 'Layer/TAUTOT', np.array)


    ####################################################################################################
 
    def calc_layering(self, H, P, T, ID, VMR, DUST, PARAH2, MOLWT=None):
        """
        Function to split atmosphere into layer and calculate their effective properties.
        
        Inputs
        ______
        
        @param H: 1D array (NP)
            Input profile heights (m)
        @param P: 1D array (NP)
            Input profile pressures (Pa)
        @param T: 1D array (NP)
            Input profile temperatures (K) 
        @param ID: 1D array (NVMR)
            Gas identifiers.
        @param VMR: 2D array (NP,NVMR)
            VMR[i,j] is Volume Mixing Ratio of gas j at vertical point i
            the column j corresponds to the gas with RADTRANS ID ID[j].
        @param DUST: 2D array (NP,NDUST)
            DUST[i,j] is abundance of the dust population j at vertical point i
            the column j corresponds to the dust population.
        @param MOLWT: 1D array (NP)
            Molecular weight at each altitude level in kg mol-1 (only required if DUST_UNITS are in particles per gram of atm)
        """
        
        #Splitting the atmosphere into layers - Calculating the BASEH and BASEP of each layer
        self.layer_split(H=H, P=P, T=T, LAYANG=self.LAYANG)
        
        #Once we have the layers, we calculate the effective properties of each layer
        self.layer_average(ID=ID, VMR=VMR, DUST=DUST, PARAH2=PARAH2, MOLWT=MOLWT)

    ####################################################################################################

    def calc_layeringg(self, H, P, T, ID, VMR, DUST, PARAH2, MOLWT=None):
        """
        
        Function to split atmosphere into layer and calculate their effective properties.
        In addition, we calculate the matrices relating the properties at each of the layers
        with respect to the properties in the input profiles.
        
        Inputs
        ______
        
        @param H: 1D array
            Input profile heights (m)
        @param P: 1D array
            Input profile pressures (Pa)
        @param T: 1D array
            Input profile temperatures (K)
        @param LAYANG: real
            Zenith angle in degrees defined at LAYHT.
        @param ID: 1D array
            Gas identifiers.
        @param VMR: 2D array
            VMR[i,j] is Volume Mixing Ratio of gas j at vertical point i
            the column j corresponds to the gas with RADTRANS ID ID[j].
        @param DUST: 2D array
            DUST[i,j] is abundance of the dust population j at vertical point i
            the column j corresponds to the dust population.
        @param MOLWT: 1D array (NP)
            Molecular weight at each altitude level in kg mol-1 (only required if DUST_UNITS are in particles per gram of atm)
        """
        
        #Splitting the atmosphere into layers - Calculating the BASEH and BASEP of each layer
        self.layer_split(H=H, P=P, T=T, LAYANG=self.LAYANG)
        
        #Once we have the layers, we calculate the effective properties of each layer
        self.layer_averageg(ID=ID, VMR=VMR, DUST=DUST, PARAH2=PARAH2, MOLWT=MOLWT)

    ####################################################################################################

    def layer_split(self, H, P, T, LAYANG):
        """
        Function to split atmosphere into layers and calculate the base altitudes and pressures
        
        Inputs
        ______
        
        @param H: 1D array
            Input profile heights (m)
        @param P: 1D array
            Input profile pressures (Pa)
        @param T: 1D array
            Input profile temperatures (K)
        @param LAYANG: real
            Zenith angle in degrees defined at LAYHT.
        """
        self.H = H
        self.P = P
        self.T = T
        self.LAYANG = LAYANG
        # get layer base height and base pressure
        BASEH, BASEP = layer_split(RADIUS=self.RADIUS, H=H, P=P, LAYANG=LAYANG,
            LAYHT=self.LAYHT, NLAY=self.NLAY, LAYTYP=self.LAYTYP,
            H_base=self.H_base, P_base=self.P_base)
        self.BASEH = BASEH
        self.BASEP = BASEP

    ####################################################################################################

    def layer_average(self, ID, VMR, DUST, PARAH2, MOLWT=None):
        """
        Function to calculate the effective properties of each layer based on the
        layer boundaries, and the reference input profiles
        
        Inputs
        ______
        
        @param ID: 1D array
            Gas identifiers.
        @param VMR: 2D array
            VMR[i,j] is Volume Mixing Ratio of gas j at vertical point i
            the column j corresponds to the gas with RADTRANS ID ID[j].
        @param DUST: 2D array
            DUST[i,j] is abundance of the dust population j at vertical point i
            the column j corresponds to the dust population.
        @param PARAH2: 1D array
            Para-H2 fraction at each altitude level
        @param MOLWT: 1D array
            Molecular weight at each altitude level 
            (only required if DUST_UNITS is in partcles per gram of atm)
        """

        HEIGHT,PRESS,TEMP,TOTAM,AMOUNT,PP,CONT,FRAC,DELH,BASET,LAYSF\
            = layer_average(RADIUS=self.RADIUS, H=self.H, P=self.P, T=self.T,
                ID=ID, VMR=VMR, DUST=DUST, PARAH2=PARAH2, BASEH=self.BASEH, BASEP=self.BASEP,
                LAYANG=self.LAYANG, LAYINT=self.LAYINT, LAYHT=self.LAYHT,
                NINT=self.NINT, DUST_UNITS = self.DUST_UNITS_FLAG, XMOLWT=MOLWT)
        self.HEIGHT = HEIGHT
        self.PRESS = PRESS
        self.TEMP = TEMP
        self.TOTAM = TOTAM
        self.AMOUNT = AMOUNT
        self.PP = PP
        self.CONT = CONT
        self.FRAC = FRAC
        self.DELH = DELH
        self.BASET = BASET
        self.LAYSF = LAYSF

    ####################################################################################################

    def layer_averageg(self, ID, VMR, DUST, PARAH2, MOLWT=None):
        """
        Function to calculate the effective properties of each layer based on the
        layer boundaries, and the reference input profiles. In addition, this function
        calculates the matrices relating the effective properties at each layer
        and the values in the input profiles.
        
        Inputs
        ______
        
        @param ID: 1D array
            Gas identifiers.
        @param VMR: 2D array
            VMR[i,j] is Volume Mixing Ratio of gas j at vertical point i
            the column j corresponds to the gas with RADTRANS ID ID[j].
        @param DUST: 2D array
            DUST[i,j] is abundance of the dust population j at vertical point i
            the column j corresponds to the dust population.
        @param MOLWT: 1D array
            Molecular weight at each altitude level 
            (only required if DUST_UNITS is in partcles per gram of atm)
        """
        # get averaged layer properties
        HEIGHT,PRESS,TEMP,TOTAM,AMOUNT,PP,CONT,FRAC,DELH,BASET,LAYSF,DTE,DAM,DCO,DPH\
            = layer_averageg(RADIUS=self.RADIUS, H=self.H, P=self.P, T=self.T,
                ID=ID, VMR=VMR, DUST=DUST, PARAH2=PARAH2, BASEH=self.BASEH, BASEP=self.BASEP,
                LAYANG=self.LAYANG, LAYINT=self.LAYINT, LAYHT=self.LAYHT,
                NINT=self.NINT, DUST_UNITS = self.DUST_UNITS_FLAG, XMOLWT=MOLWT)
        self.HEIGHT = HEIGHT
        self.PRESS = PRESS
        self.TEMP = TEMP
        self.TOTAM = TOTAM
        self.AMOUNT = AMOUNT
        self.PP = PP
        self.CONT = CONT
        self.FRAC = FRAC
        self.DELH = DELH
        self.BASET = BASET
        self.LAYSF = LAYSF
        self.DTE = DTE
        self.DAM = DAM
        self.DCO = DCO
        self.DPH = DPH
    ####################################################################################################

    def plot_Layer(self):
        """
        Make a summary plot summarising the information about the layers in the atmosphere
        """
        
        #Making a summary plot 
        fig,ax = plt.subplots(1,4,figsize=(10,3),sharey=True)

        #Pressure
        for i in range(self.NLAY):
            ax[0].axhline(self.BASEH[i]/1.0e3,c='black',linestyle='--',linewidth=0.5)
        ax[0].plot(self.PRESS/1.0e5,self.HEIGHT/1.0e3,marker='o',markersize=4.,c='tab:blue')
        ax[0].set_xscale('log')
        ax[0].set_xlabel('Effective pressure (bar)')
        ax[0].set_ylabel('Altitude (km)')
        ax[0].set_facecolor('lightgray')

        #Temperature
        for i in range(self.NLAY):
            ax[1].axhline(self.BASEH[i]/1.0e3,c='black',linestyle='--',linewidth=0.5)
        ax[1].plot(self.TEMP,self.HEIGHT/1.0e3,marker='o',markersize=4.,c='tab:red')
        ax[1].set_xlabel('Effective temperature (K)')
        #ax[1].set_ylabel('Altitude (km)')
        ax[1].set_facecolor('lightgray')

        #Gaseous column density
        for i in range(self.NLAY):
            ax[2].axhline(self.BASEH[i]/1.0e3,c='black',linestyle='--',linewidth=0.5)
        for i in range(self.PP.shape[1]):
            ax[2].plot(self.AMOUNT[:,i],self.HEIGHT/1.0e3,marker='o',markersize=4.)
        ax[2].set_xscale('log')
        ax[2].set_xlabel('Gaseous column density (m$^{-2}$)')
        #ax[2].set_ylabel('Altitude (km)')
        ax[2].set_facecolor('lightgray')

        #Dust column density
        for i in range(self.NLAY):
            ax[3].axhline(self.BASEH[i]/1.0e3,c='black',linestyle='--',linewidth=0.5)
        for i in range(self.CONT.shape[1]):
            ax[3].plot(self.CONT[:,i],self.HEIGHT/1.0e3,marker='o',markersize=4.)
        ax[3].set_xscale('log')
        ax[3].set_xlabel('Aerosol column density (m$^{-2}$)')
        #ax[3].set_ylabel('Altitude (km)')
        ax[3].set_facecolor('lightgray')

        plt.tight_layout()


#########################################################################################
# END OF LAYER CLASS

# USEFUL FUNCTIONS FOR THE LAYERING CLASS
#########################################################################################

def interp(X_data, Y_data, X, ITYPE=InterpolationMethod.LINEAR):
    """
    Routine for 1D interpolation using the SciPy library.

    Inputs
    ------
    @param X_data: 1D array

    @param Y_data: 1D array

    @param X: real

    @param ITYPE: InterpolationMethod (enum)
        InterpolationMethod.LINEAR=linear interpolation
        InterpolationMethod.QUADRATIC_SPLINE=quadratic spline interpolation
        InterpolationMethod.CUBIC_SPLINE=cubic spline interpolation
    """
    
    from scipy.interpolate import interp1d
    
    if ITYPE == InterpolationMethod.LINEAR:
        f = interp1d(X_data, Y_data, kind='linear', fill_value='extrapolate')
        Y = f(X)

    elif ITYPE == InterpolationMethod.QUADRATIC_SPLINE:
        f = interp1d(X_data, Y_data, kind='quadratic', fill_value='extrapolate')
        Y = f(X)

    elif ITYPE == InterpolationMethod.CUBIC_SPLINE:
        f = interp1d(X_data, Y_data, kind='cubic', fill_value='extrapolate')
        Y = f(X)

    return Y

#########################################################################################

@jit(nopython=True)
def interp_numba(X_data, Y_data, X):
    """
    Optimised routine for 1D linear interpolation using Numba JIT.

    Inputs
    ------
    @param X_data: 1D array (monotonic increasing values)
    @param Y_data: 1D array
    @param X: 1D array (values to interpolate)

    Returns
    -------
    @return Y: Interpolated values at X
    @return J: Indices of the left bounding points
    @return F: Fractional distances between interpolation points
    """

    # Ensure X is always an array
    instance=False
    if isinstance(X, (float, np.float64, int)):
        instance=True
        X = np.array([X])  # Convert single value to array
        
    if np.any(np.diff(X_data) <= 0):
        raise ValueError("error in interp :: X_data must be strictly increasing.")
        
    NX = len(X)
    Y = np.zeros(NX)
    J = np.zeros(NX, dtype=np.int32)
    F = np.zeros(NX)

    for IX in range(NX):
        j = 0
        while j < len(X_data) and X_data[j] <= X[IX]:
            j += 1
        j = min(j, len(X_data) - 1)  # Ensure j does not exceed bounds

        if j == 0:
            j = 1  # Avoid out-of-bounds access

        J[IX] = j - 1
        F[IX] = (X[IX] - X_data[j - 1]) / (X_data[j] - X_data[j - 1])
        Y[IX] = (1.0 - F[IX]) * Y_data[j - 1] + F[IX] * Y_data[j]

    if instance==True:
        return Y[0] 
    else:
        return Y

#########################################################################################

@jit(nopython=True)
def interpg(X_data, Y_data, X):
    """
    Optimised routine for 1D linear interpolation using Numba JIT.

    Inputs
    ------
    @param X_data: 1D array (monotonic increasing values)
    @param Y_data: 1D array
    @param X: 1D array (values to interpolate)

    Returns
    -------
    @return Y: Interpolated values at X
    @return J: Indices of the left bounding points
    @return F: Fractional distances between interpolation points
    """

    NX = len(X)
    Y = np.zeros(NX)
    J = np.zeros(NX, dtype=np.int32)
    F = np.zeros(NX)

    for IX in range(NX):
        j = 0
        while j < len(X_data) and X_data[j] <= X[IX]:
            j += 1
        j = min(j, len(X_data) - 1)  # Ensure j does not exceed bounds

        if j == 0:
            j = 1  # Avoid out-of-bounds access

        J[IX] = j - 1
        F[IX] = (X[IX] - X_data[j - 1]) / (X_data[j] - X_data[j - 1])
        Y[IX] = (1.0 - F[IX]) * Y_data[j - 1] + F[IX] * Y_data[j]

    return Y, J, F

#########################################################################################

def layer_average(RADIUS, H, P, T, ID, VMR, DUST, PARAH2, BASEH, BASEP,
                  LAYANG=0.0, LAYINT=LayerIntegrationScheme.MID_PATH, LAYHT=0.0, NINT=101, DUST_UNITS=None, XMOLWT=None):
    """
    Calculates average layer properties.
    Takes an atmosphere profile and a layering shceme specified by
    layer base altitudes and pressures and returns averaged height,
    pressure, temperature, VMR for each layer.

    Inputs
    ------
    @param RADIUS: real
        Reference planetary radius where H=0.  Usually at surface for
        terrestrial planets, or at 1 bar pressure level for gas giants.
    @param H: 1D array
        Input profile heights
    @param P: 1D array
        Input profile pressures
    @param T: 1D array
        Input profile temperatures
    @param ID: 1D array
        Gas identifiers.
    @param VMR: 2D array
        VMR[i,j] is Volume Mixing Ratio of gas j at vertical point i
        the column j corresponds to the gas with RADTRANS ID ID[j].
    @param DUST: 2D array
        DUST[i,j] is dust density of dust popoulation j at vertical point i  (particles/m3)
    @param PARAH2: 1D array
        Para-H2 fraction
    @param BASEH: 1D array
        Heights of the layer bases.
    @param BASEP: 1D array
        Pressures of the layer bases.
    @param LAYANG: real
        Zenith angle in degrees defined at LAYHT.
    @param LAYINT: LayerIntegrationScheme (enum)
        Layer integration scheme
        LayerIntegrationScheme.MID_PATH = use properties at mid path
        LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE = use absorber amount weighted average values
    @param LAYHT: real
        Height of the base of the lowest layer. Default 0.0.
    @param NINT: int
        Number of integration points to be used if LAYINT=LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE.
    @param XMOLWT: 1D array
        Molecular weight vertical profile in kg mol-1 (only required if DUST_UNITS is activated)

    Returns
    -------
    @param HEIGHT: 1D array
        Representative height for each layer
    @param PRESS: 1D array
        Representative pressure for each layer
    @param TEMP: 1D array
        Representative pressure for each layer
    @param TOTAM: 1D array
        Total gaseous absorber amounts along the line-of-sight path, i.e.
        number of molecules per area.
    @param AMOUNT: 1D array
        Representative absorber amounts of each gas at each layer.
        AMOUNT[I,J] is the representative number of gas J molecules
        in layer I in the form of number of molecules per area.
    @param PP: 1D array
        Representative partial pressure for each gas at each layer.
        PP[I,J] is the representative partial pressure of gas J in layer I.
    @param DELH: 1D array
        Layer thicnkness.
    @param BASET: 1D array
        Layer base temperature.
    @param LAYSF: 1D array
        Layer scaling factor.
    """

    from scipy.integrate import simpson
    
    k_B = 1.38065e-23

    # Calculate layer geometric properties
    NLAY = len(BASEH)
    NPRO = len(H)
    DELH = np.zeros(NLAY)           # layer thickness
    DELH[0:-1] = (BASEH[1:]-BASEH[:-1])
    DELH[-1] = (H[-1]-BASEH[-1])

    sin = np.sin(LAYANG*np.pi/180)  # sin(viewing angle)
    cos = np.cos(LAYANG*np.pi/180)  # cos(viewing angle)

    z0 = RADIUS + LAYHT    # distance from planet centre to lowest layer base
    zmax = RADIUS+H[-1]    # maximum height defined in the profile

    SMAX = np.sqrt(zmax**2-(z0*sin)**2)-z0*cos # total path length
    BASES = np.sqrt((RADIUS+BASEH)**2-(z0*sin)**2)-z0*cos # Path lengths at base of layer
    DELS = np.zeros(NLAY)           # path length in each layer
    DELS[0:-1] = (BASES[1:]-BASES[:-1])
    DELS[-1] = (SMAX-BASES[-1])
    LAYSF = DELS/DELH               # Layer Scaling Factor
    BASET = interp(H,T,BASEH)       # layer base temperature

    # Note number of: profile points, layers, VMRs, aerosol types, flags
    if VMR.ndim == 1:
        NVMR = 1
    else:
        NVMR = len(VMR[0])

    if DUST.ndim == 1:
        NDUST = 1
    else:
        NDUST = len(DUST[0])

    #Checking consistency of DUST_UNITS and XMOLWT
    if DUST_UNITS is not None:
        for j in range(NDUST):
            if DUST_UNITS[j]==-1:
                if XMOLWT is None:
                    raise ValueError('error in layer_average :: if DUST_UNITS=-1 (particles per gram of atm), the XMOLWT must be defined')
        XMOLWT *= 1000. #Converting to molecular weight in g mol-1
    else:
        XMOLWT = np.zeros(NPRO)
    
    # HEIGHT = average layer height
    HEIGHT = np.zeros(NLAY)
    # PRESS = average layer pressure
    PRESS  = np.zeros(NLAY)
    # TEMP = average layer temperature
    TEMP   = np.zeros(NLAY)
    # TOTAM = total no. of molecules/aera
    TOTAM  = np.zeros(NLAY)
    # DUDS = no. of molecules per area per distance
    DUDS   = np.zeros(NLAY)
    # AMOUNT = no. of molecules/aera for each gas (molecules/m2)
    AMOUNT = np.zeros((NLAY, NVMR))
    # PP = gas partial pressures
    PP     = np.zeros((NLAY, NVMR))
    # MOLWT = mean molecular weight
    MOLWT  = np.zeros(NLAY)
    # CONT = no. of particles/area for each dust population (particles/m2)
    CONT = np.zeros((NLAY,NDUST))
    
    # FRAC = fraction of para-H2 in each layer
    FRAC   = np.zeros(NLAY)
    if PARAH2 is None:
        PARAH2 = np.zeros(NPRO)

    # Calculate average properties depending on intergration type
    if LAYINT == LayerIntegrationScheme.MID_PATH:
        # use layer properties at half path length in each layer
        S = np.zeros(NLAY)
        S[:-1] = (BASES[1:] + BASES[:-1])/2
        S[-1] = (SMAX+BASES[-1])/2
        # Derive other properties from S
        HEIGHT = np.sqrt(S**2+z0**2+2*S*z0*cos) - RADIUS
        PRESS = interp(H,P,HEIGHT)
        TEMP = interp(H,T,HEIGHT)
        FRAC = interp(H,PARAH2,HEIGHT)
        MOLWT = interp(H,XMOLWT,HEIGHT)
        
        # Ideal gas law: N/(Area*Path_length) = P/(k_B*T)
        DUDS = PRESS/(k_B*TEMP)
        TOTAM = DUDS*DELS
        # Use the volume mixing ratio information
        if VMR.ndim > 1:
            AMOUNT = np.zeros((NLAY, NVMR))
            for J in range(NVMR):
                AMOUNT[:,J] = interp(H, VMR[:,J], HEIGHT)
            PP = (AMOUNT.T * PRESS).T
            AMOUNT = (AMOUNT.T * TOTAM).T
        else:
            AMOUNT = interp(H, VMR, HEIGHT)
            PP = AMOUNT * PRESS
            AMOUNT = AMOUNT * TOTAM

        #Use the dust density information
        if DUST.ndim > 1:
            for J in range(NDUST):
                DD = interp(H, DUST[:,J],HEIGHT)  
                if DUST_UNITS is not None:
                    if DUST_UNITS[J] == -1:  #dust units are particles per gram of atm
                        CONT[:,J] =  DD * TOTAM * MOLWT / AVOGAD
                    else:
                         CONT[:,J] = DD * DELS  
                else:  
                    CONT[:,J] = DD * DELS
        else:
            DD = interp(H, DUST,HEIGHT)  
            if DUST_UNITS is not None:
                if DUST_UNITS[0] == -1: #dust units are particles per gram of atm
                    CONT =  DD * TOTAM * MOLWT / AVOGAD
                else:
                    CONT = DD * DELS
            else:
                CONT = DD * DELS

    elif LAYINT == LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE:
        # Curtis-Godson equivalent path for a gas with constant mixing ratio
        for I in range(NLAY):
            S0 = BASES[I]
            if I < NLAY-1:
                S1 = BASES[I+1]
            else:
                S1 = SMAX
            # sub-divide each layer into NINT layers
            S = np.linspace(S0, S1, NINT)
            h = np.sqrt(S**2+z0**2+2*S*z0*cos)-RADIUS
            p = interp(H,P,h)
            temp = interp(H,T,h)
            frac = interp(H,PARAH2,h)
            molwt = interp(H,XMOLWT,h)
            
            duds = p/(k_B*temp)

            amount = np.zeros((NINT, NVMR))
            
            TOTAM[I] = simpson(duds,x=S)
            HEIGHT[I]  = simpson(h*duds,x=S)/TOTAM[I]
            PRESS[I] = simpson(p*duds,x=S)/TOTAM[I]
            TEMP[I]  = simpson(temp*duds,x=S)/TOTAM[I]
            FRAC[I]  = simpson(frac*duds,x=S)/TOTAM[I]
            MOLWT[I] = simpson(molwt*duds,x=S)/TOTAM[I]
            
            if VMR.ndim > 1:
                amount = np.zeros((NINT, NVMR))
                for J in range(NVMR):
                    amount[:,J] = interp(H, VMR[:,J], h)
                    AMOUNT[I,J] = simpson(amount[:,J]*duds,x=S)
                pp = (amount.T * p).T     # gas partial pressures
                for J in range(NVMR):
                    PP[I, J] = simpson(pp[:,J]*duds,x=S)/TOTAM[I]
            else:
                amount = interp(H, VMR, h)
                pp = amount * p
                AMOUNT[I] = simpson(amount*duds,x=S)
                PP[I] = simpson(pp*duds,x=S)/TOTAM[I]
                        
            if DUST.ndim > 1:
                dd = np.zeros((NINT,NDUST))
                for J in range(NDUST):
                    dd[:,J] = interp(H, DUST[:,J], h)
                    if DUST_UNITS is not None:
                        if DUST_UNITS[J] == -1: #dust units are particles per gram of atm
                            CONT[I,J] = simpson(dd[:,J]*duds*molwt/AVOGAD,x=S)
                        else:
                            CONT[I,J] = simpson(dd[:,J],x=S)
                    else:
                        CONT[I,J] = simpson(dd[:,J],x=S)     
            else:
                dd = interp(H, DUST, h) 
                if DUST_UNITS is not None:
                    if DUST_UNITS[0] == -1: #dust units are particles per gram of atm
                        CONT[I] = simpson(dd*duds*molwt/AVOGAD,x=S)
                    else:
                        CONT[I] = simpson(dd,x=S)
                else:
                    CONT[I] = simpson(dd,x=S)
            
    # Scale back to vertical layers
    TOTAM = TOTAM / LAYSF
    if VMR.ndim > 1:
        AMOUNT = (AMOUNT.T * LAYSF**-1 ).T
    else:
        AMOUNT = AMOUNT/LAYSF

    if DUST.ndim > 1:
        CONT = (CONT.T * LAYSF**-1 ).T
    else:
        CONT = CONT/LAYSF
        
    if DUST_UNITS is not None:
        XMOLWT /= 1000. #Converting back to molecular weight in kg mol-1
        
    return HEIGHT,PRESS,TEMP,TOTAM,AMOUNT,PP,CONT,FRAC,DELH,BASET,LAYSF

#########################################################################################

def layer_averageg(RADIUS, H, P, T, ID, VMR, DUST, PARAH2, BASEH, BASEP,
                  LAYANG=0.0, LAYINT=LayerIntegrationScheme.MID_PATH, LAYHT=0.0, NINT=101, DUST_UNITS=None, XMOLWT=None):
    """
    Calculates average layer properties.
    Takes an atmosphere profile and a layering shceme specified by
    layer base altitudes and pressures and returns averaged height,
    pressure, temperature, VMR for each layer.

    Inputs
    ------
    @param RADIUS: real
        Reference planetary radius where H=0.  Usually at surface for
        terrestrial planets, or at 1 bar pressure level for gas giants.
    @param H: 1D array
        Input profile heights
    @param P: 1D array
        Input profile pressures
    @param T: 1D array
        Input profile temperatures
    @param ID: 1D array
        Gas identifiers.
    @param VMR: 2D array
        VMR[i,j] is Volume Mixing Ratio of gas j at vertical point i
        the column j corresponds to the gas with RADTRANS ID ID[j].
    @param DUST: 2D array
        DUST[i,j] is dust density of dust popoulation j at vertical point i  (particles/m3)
    @param PARAH2: 1D array
        Para-H2 fraction
    @param BASEH: 1D array
        Heights of the layer bases.
    @param BASEP: 1D array
        Pressures of the layer bases.
    @param LAYANG: real
        Zenith angle in degrees defined at LAYHT.
    @param LAYINT: LayerIntegrationScheme (enum)
        Layer integration scheme
        LayerIntegrationScheme.MID_PATH = use properties at mid path
        LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE = use absorber amount weighted average values
    @param LAYHT: real
        Height of the base of the lowest layer. Default 0.0.
    @param NINT: int
        Number of integration points to be used if LAYINT=LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE.
    @param XMOLWT: 1D array
        Molecular weight vertical profile in kg mol-1 (only required if DUST_UNITS is activated)

    Returns
    -------
    @param HEIGHT: 1D array
        Representative height for each layer
    @param PRESS: 1D array
        Representative pressure for each layer
    @param TEMP: 1D array
        Representative pressure for each layer
    @param TOTAM: 1D array
        Total gaseous absorber amounts along the line-of-sight path, i.e.
        number of molecules per area.
    @param AMOUNT: 1D array
        Representative absorber amounts of each gas at each layer.
        AMOUNT[I,J] is the representative number of gas J molecules
        in layer I in the form of number of molecules per area.
    @param PP: 1D array
        Representative partial pressure for each gas at each layer.
        PP[I,J] is the representative partial pressure of gas J in layer I.
    @param DELH: 1D array
        Layer thicnkness.
    @param BASET: 1D array
        Layer base temperature.
    @param LAYSF: 1D array
        Layer scaling factor.
    @param DTE: 2D array
        Matrix relating the temperature in each layer (TEMP) to the input temperature profile (T)
    @param DAM: 2D array
        Matrix relating the absorber amounts in each layer (AMOUNT) to the input VMR profile (VMR)
    @param DCO: 2D array
        Matrix relating the dust amounts in each layer (CONT) to the input dust profile (DUST)
    """

    from scipy.integrate import simpson

    k_B = 1.38065e-23

    # Calculate layer geometric properties
    NLAY = len(BASEH)
    NPRO = len(H)
    DELH = np.zeros(NLAY)           # layer thickness
    DELH[0:-1] = (BASEH[1:]-BASEH[:-1])
    DELH[-1] = (H[-1]-BASEH[-1])

    sin = np.sin(LAYANG*np.pi/180)  # sin(viewing angle)
    cos = np.cos(LAYANG*np.pi/180)  # cos(viewing angle)

    z0 = RADIUS + LAYHT    # distance from planet centre to lowest layer base
    zmax = RADIUS+H[-1]    # maximum height defined in the profile

    SMAX = np.sqrt(zmax**2-(z0*sin)**2)-z0*cos # total path length
    BASES = np.sqrt((RADIUS+BASEH)**2-(z0*sin)**2)-z0*cos # Path lengths at base of layer
    DELS = np.zeros(NLAY)           # path length in each layer
    DELS[0:-1] = (BASES[1:]-BASES[:-1])
    DELS[-1] = (SMAX-BASES[-1])
    LAYSF = DELS/DELH               # Layer Scaling Factor
    BASET = interp(H,T,BASEH)       # layer base temperature

    # Note number of: profile points, layers, VMRs, aerosol types, flags
    if VMR.ndim == 1:
        NVMR = 1
    else:
        NVMR = len(VMR[0])

    if DUST.ndim == 1:
        NDUST = 1
    else:
        NDUST = len(DUST[0])

    #Checking consistency of DUST_UNITS and XMOLWT
    if DUST_UNITS is not None:
        for j in range(NDUST):
            if DUST_UNITS[j]==-1:
                if XMOLWT is None:
                    raise ValueError('error in layer_average :: if DUST_UNITS=-1 (particles per gram of atm), the XMOLWT must be defined')
        XMOLWT *= 1000. #Converting to molecular weight in g mol-1
    else:
        XMOLWT = np.zeros(NPRO)

    # HEIGHT = average layer height
    HEIGHT = np.zeros(NLAY)
    # PRESS = average layer pressure
    PRESS  = np.zeros(NLAY)
    # TEMP = average layer temperature
    TEMP   = np.zeros(NLAY)
    # TOTAM = total no. of molecules/aera
    TOTAM  = np.zeros(NLAY)
    # DUDS = no. of molecules per area per distance
    DUDS   = np.zeros(NLAY)
    # AMOUNT = no. of molecules/aera for each gas (molecules/m2)
    AMOUNT = np.zeros((NLAY, NVMR))
    # PP = gas partial pressures
    PP     = np.zeros((NLAY, NVMR))
    # MOLWT = mean molecular weight
    MOLWT  = np.zeros(NLAY)
    # CONT = no. of particles/area for each dust population (particles/m2)
    CONT = np.zeros((NLAY,NDUST))
    # FRAC = fraction of para-H2 in each layer
    FRAC = np.zeros(NLAY)
    if PARAH2 is None:
        PARAH2 = np.zeros(NPRO)
    #DTE = matrix to relate the temperature in each layer (TEMP) to the temperature in the input profiles (T)
    DTE = np.zeros((NLAY, NPRO))
    #DCO = matrix to relate the dust abundance in each layer (CONT) to the dust abundance in the input profiles (DUST)
    DCO = np.zeros((NLAY, NPRO))
    #DAM = matrix to relate the gaseous abundance in each layer (AMOUNT) to the gas VMR in the input profiles (VMR)
    DAM = np.zeros((NLAY, NPRO))
    #DPH = matrix to relate para-H2 in each layer (FRAC) to that in the input profiles (PARAH2)
    DPH = np.zeros((NLAY, NPRO))

    # Ensure NINT is odd for Simpson's rule
    if NINT % 2 == 0:
        raise ValueError("NINT must be odd for Simpson's rule.")
    # Initialize weights for Simpson's rule
    w = np.ones(NINT)  # Start with all ones
    # Set 4s for odd indices (1-based, so 1,3,5,... -> 0-based 1,3,5,...)
    w[1:-1:2] = 4.
    # Set 2s for even indices (1-based 2,4,6,... -> 0-based 2,4,6,...)
    w[2:-1:2] = 2.

    # Calculate average properties depending on intergration type
    if LAYINT == LayerIntegrationScheme.MID_PATH:
        # use layer properties at half path length in each layer
        S = np.zeros(NLAY)
        S[:-1] = (BASES[1:] + BASES[:-1])/2
        S[-1] = (SMAX+BASES[-1])/2
        # Derive other properties from S
        HEIGHT = np.sqrt(S**2+z0**2+2*S*z0*cos) - RADIUS
        PRESS = interp(H,P,HEIGHT)
        MOLWT = interp(H,XMOLWT,HEIGHT)
        TEMP,J,F = interpg(H,T,HEIGHT)
        for ilay in range(NLAY):
            DTE[ilay,J[ilay]] = DTE[ilay,J[ilay]] + (1.0-F[ilay])
            DTE[ilay,J[ilay]+1] = DTE[ilay,J[ilay]+1] + (F[ilay])

        FRAC,J,F = interpg(H,PARAH2,HEIGHT)
        for ilay in range(NLAY):
            DPH[ilay,J[ilay]] = DPH[ilay,J[ilay]] + (1.0-F[ilay])
            DPH[ilay,J[ilay]+1] = DPH[ilay,J[ilay]+1] + (F[ilay])
            
        # Ideal gas law: N/(Area*Path_length) = P/(k_B*T)
        DUDS = PRESS/(k_B*TEMP) 
        TOTAM = DUDS*DELS
        # Use the volume mixing ratio information
        if VMR.ndim > 1:
            AMOUNT = np.zeros((NLAY, NVMR))
            for J in range(NVMR):
                AMOUNT[:,J],JJ,F = interpg(H, VMR[:,J], HEIGHT)
            PP = (AMOUNT.T * PRESS).T
            AMOUNT = (AMOUNT.T * TOTAM).T
        else:
            AMOUNT,JJ,F = interpg(H, VMR, HEIGHT)
            PP = AMOUNT * PRESS
            AMOUNT = AMOUNT * TOTAM

        for ilay in range(NLAY):
            DAM[ilay,JJ[ilay]] = DAM[ilay,JJ[ilay]] + (1.0-F[ilay])*TOTAM[ilay]
            DAM[ilay,JJ[ilay]+1] = DAM[ilay,JJ[ilay]+1] + (F[ilay])*TOTAM[ilay]

        #Use the dust density information
        if DUST.ndim > 1:
            for J in range(NDUST):
                DD,JJ,F = interpg(H, DUST[:,J],HEIGHT)  
                CONT[:,J] = DD * DELS
        else:
            #DD = interp(H, DUST,HEIGHT)  
            DD,JJ,F = interpg(H, DUST,HEIGHT)
            CONT = DD * DELS            


        #Use the dust density information
        dust_units_flag = False
        if DUST.ndim > 1:
            for J in range(NDUST):
                DD = interp(H, DUST[:,J],HEIGHT)  
                if DUST_UNITS is not None:
                    if DUST_UNITS[J] == -1:  #dust units are particles per gram of atm
                        dust_units_flag = True
                        CONT[:,J] =  DD * TOTAM * MOLWT / AVOGAD
                    else:
                         CONT[:,J] = DD * DELS  
                else:  
                    CONT[:,J] = DD * DELS
        else:
            DD = interp(H, DUST,HEIGHT)  
            if DUST_UNITS is not None:
                if DUST_UNITS[0] == -1: #dust units are particles per gram of atm
                    dust_units_flag = True
                    CONT =  DD * TOTAM * MOLWT / AVOGAD
                else:
                    CONT = DD * DELS
            else:
                CONT = DD * DELS

        if dust_units_flag is False:
            for ilay in range(NLAY):
                DCO[ilay,JJ[ilay]] = DCO[ilay,JJ[ilay]] + (1.0-F[ilay])
                DCO[ilay,JJ[ilay]+1] = DCO[ilay,JJ[ilay]+1] + (F[ilay])
        else:
                DCO[ilay,JJ[ilay]] = DCO[ilay,JJ[ilay]] + (1.0-F[ilay]) * TOTAM * MOLWT /AVOGAD
                DCO[ilay,JJ[ilay]+1] = DCO[ilay,JJ[ilay]+1] + (F[ilay]) * TOTAM * MOLWT /AVOGAD
            
    elif LAYINT == LayerIntegrationScheme.ABSORBER_WEIGHTED_AVERAGE:
        
        # Curtis-Godson equivalent path for a gas with constant mixing ratio
        for I in range(NLAY):
            S0 = BASES[I]
            if I < NLAY-1:
                S1 = BASES[I+1]
            else:
                S1 = SMAX
            # sub-divide each layer into NINT layers
            S = np.linspace(S0, S1, NINT)
            h = np.sqrt(S**2+z0**2+2*S*z0*cos)-RADIUS   #altitude of the NINT points (m)
            p = interp(H,P,h)                           #pressure at the NINT points (Pa)
            molwt = interp(H,XMOLWT,h)                           #molecular weight at the NINT points (g mol-1)
            temp,JJ,F = interpg(H,T,h)                  #temperature at the NINT points (K)
            frac,JJP,FP = interpg(H,PARAH2,h)           #fraction of para-H2 at the NINT points 
            duds = p/(k_B*temp)                         #number density at the NINT points (m-3)

            amount = np.zeros((NINT, NVMR))

            TOTAM[I] = simpson(duds,x=S)                #column density (m-2)
            HEIGHT[I]  = simpson(h*duds,x=S)/TOTAM[I]   #effective altitude of the layer (m)
            PRESS[I] = simpson(p*duds,x=S)/TOTAM[I]     #effective pressure of the layer (Pa)
            TEMP[I]  = simpson(temp*duds,x=S)/TOTAM[I]  #effective temperature of the layer (K)
            FRAC[I]  = simpson(frac*duds,x=S)/TOTAM[I]  #effective fraction of para-H2 in the layer

            for iint in range(NINT):
                DTE[I,JJ[iint]] = DTE[I,JJ[iint]] + (1.-F[iint])*w[iint]*duds[iint]
                DTE[I,JJ[iint]+1] = DTE[I,JJ[iint]+1] + (F[iint])*w[iint]*duds[iint]

                DPH[I,JJP[iint]] = DPH[I,JJP[iint]] + (1.-FP[iint])*w[iint]*duds[iint]
                DPH[I,JJP[iint]+1] = DPH[I,JJP[iint]+1] + (FP[iint])*w[iint]*duds[iint]

                
            #Computing the gas column densities and gradients
            if VMR.ndim > 1:
                amount = np.zeros((NINT, NVMR))
                for J in range(NVMR):
                    amount[:,J],JJ,F = interpg(H, VMR[:,J], h)    #VMR of gas J at the NINT points
                    AMOUNT[I,J] = simpson(amount[:,J]*duds,x=S)   #column density of gas J in the layer (m-2)
                pp = (amount.T * p).T     # gas partial pressures at the NINT points
                for J in range(NVMR):
                    PP[I, J] = simpson(pp[:,J]*duds,x=S)/TOTAM[I]   #effective partial pressure of gas J in the layer
            else:
                amount,JJ,F = interpg(H, VMR, h)   
                pp = amount * p
                AMOUNT[I] = simpson(amount*duds,x=S)
                PP[I] = simpson(pp*duds,x=S)/TOTAM[I]

            for iint in range(NINT):
                DAM[I,JJ[iint]] = DAM[I,JJ[iint]] + (1.-F[iint])*duds[iint]*w[iint]
                DAM[I,JJ[iint]+1] = DAM[I,JJ[iint]+1] + (F[iint])*duds[iint]*w[iint]


            #Computing the aerosol column densities and gradients
            dust_units_flag = False
            if DUST.ndim > 1:
                dd = np.zeros((NINT,NDUST))
                for J in range(NDUST):
                    dd[:,J],JJ,F = interpg(H, DUST[:,J], h)   #dust density (m-3) of dust population J at the NINT points
                    if DUST_UNITS is not None:
                        if DUST_UNITS[J] == -1: #dust units are in particles per gram of atm
                            dust_units_flag = True
                            CONT[I,J] = simpson(dd[:,J]*duds*molwt/AVOGAD,x=S)  #column density of dust population J in the layer (m-2)
                        else:
                            CONT[I,J] = simpson(dd[:,J],x=S)          #column density of dust population J in the layer (m-2)
                    else:
                        CONT[I,J] = simpson(dd[:,J],x=S)          #column density of dust population J in the layer (m-2)
            else:
                dd,JJ,F = interpg(H, DUST, h) 
                if DUST_UNITS is not None:
                    if DUST_UNITS[0] == -1: #dust units are in particles per gram of atm
                        dust_units_flag = True
                        CONT[I] = simpson(dd*duds*molwt/AVOGAD,x=S)
                    else:
                        CONT[I] = simpson(dd,x=S)
                else:
                    CONT[I] = simpson(dd,x=S)
                
            if dust_units_flag is False:
                for iint in range(NINT):
                    DCO[I,JJ[iint]] = DCO[I,JJ[iint]] + (1.-F[iint])*w[iint]
                    DCO[I,JJ[iint]+1] = DCO[I,JJ[iint]+1] + (F[iint])*w[iint]
            else:
                for iint in range(NINT):
                    DCO[I,JJ[iint]] = DCO[I,JJ[iint]] + (1.-F[iint])*w[iint]*duds[iint]*molwt[iint]/AVOGAD
                    DCO[I,JJ[iint]+1] = DCO[I,JJ[iint]+1] + (F[iint])*w[iint]*duds[iint]*molwt[iint]/AVOGAD

        #Finishing the integration for the matrices
        for IPRO in range(NPRO):
            DTE[:,IPRO] = DTE[:,IPRO] * DELS / (NINT-1.) / 3. / TOTAM
            DPH[:,IPRO] = DPH[:,IPRO] * DELS / (NINT-1.) / 3. / TOTAM
            DAM[:,IPRO] = DAM[:,IPRO] * DELS / (NINT-1.) / 3.
            DCO[:,IPRO] = DCO[:,IPRO] * DELS / (NINT-1.) / 3.

    # Scale back to vertical layers
    TOTAM = TOTAM / LAYSF
    if VMR.ndim > 1:
        AMOUNT = (AMOUNT.T * LAYSF**-1 ).T
    else:
        AMOUNT = AMOUNT/LAYSF

    if DUST.ndim > 1:
        CONT = (CONT.T * LAYSF**-1 ).T
    else:
        CONT = CONT/LAYSF

    for IPRO in range(NPRO):
        DAM[:,IPRO] = DAM[:,IPRO] / LAYSF
        DCO[:,IPRO] = DCO[:,IPRO] / LAYSF

    if DUST_UNITS is not None:
        XMOLWT /= 1000. #Converting back to molecular weight in kg mol-1

    return HEIGHT,PRESS,TEMP,TOTAM,AMOUNT,PP,CONT,FRAC,DELH,BASET,LAYSF,DTE,DAM,DCO,DPH

#########################################################################################

def layer_split(RADIUS, H, P, LAYANG=0.0, LAYHT=0.0, NLAY=20,
        LAYTYP=LayerType.EQUAL_LOG_PRESSURE, H_base=None, P_base=None):
    """
    Splits an atmosphere into NLAY layers.
    Takes a set of altitudes H with corresponding pressures P and returns
    the altitudes and pressures of the base of the layers.

    Inputs
    ------
    @param RADIUS: real
        Reference planetary radius where H=0.  Usually at surface for
        terrestrial planets, or at 1 bar pressure level for gas giants.
    @param H: 1D array
        Heights at which the atmosphere profile is specified.
        (At altitude H[i] the pressure is P[i].)
    @param P: 1D array
        Pressures at which the atmosphere profile is specified.
        (At pressure P[i] the altitude is H[i].)
    @param LAYANG: real
        Zenith angle in degrees defined at LAYHT.
        Default 0.0 (nadir geometry). Only needed for layer type LayerType.EQUAL_PATH_LENGTH.
    @param LAYHT: real
        Height of the base of the lowest layer. Default 0.0.
    @param NLAY: int
        Number of layers to split the atmosphere into. Default 20.
    @param LAYTYP: LayerType (enum)
        Integer specifying how to split up the layers. Default LayerType.EQUAL_LOG_PRESSURE.
        LayerType.EQUAL_PRESSURE = by equal changes in pressure
        LayerType.EQUAL_LOG_PRESSURE = by equal changes in log pressure
        LayerType.EQUAL_HEIGHT = by equal changes in height
        LayerType.EQUAL_PATH_LENGTH = by equal changes in path length at LAYANG
        LayerType.BASE_PRESSURE = layer base pressure levels specified by P_base
        LayerType.BASE_HEIGHT = layer base height levels specified by H_base
        Note BASE_PRESSURE and BASE_HEIGHT force NLAY = len(P_base) or len(H_base).
    @param H_base: 1D array
        Heights of the layer bases defined by user. Default None.
    @param P_base: 1D array
        Pressures of the layer bases defined by user. Default None.

    Returns
    -------
    @param BASEH: 1D array
        Heights of the layer bases.
    @param BASEP: 1D array
        Pressures of the layer bases.
    """

    if LAYHT<H[0]:
        _lgr.warning(' from layer_split() :: LAYHT < H(0). Resetting LAYHT')
        LAYHT = H[0]

    if LAYTYP == LayerType.EQUAL_PRESSURE: # split by equal pressure intervals
        PBOT = interp(H,P,LAYHT)  # pressure at base of lowest layer
        BASEP = np.linspace(PBOT,P[-1],NLAY+1)[:-1]
        BASEH = interp(P[::-1],H[::-1],BASEP)

    elif LAYTYP == LayerType.EQUAL_LOG_PRESSURE: # split by equal log pressure intervals
        PBOT = interp(H,P,LAYHT)  # pressure at base of lowest layer
        BASEP = np.logspace(np.log10(PBOT),np.log10(P[-1]),NLAY+1)[:-1]
        BASEH = interp(P[::-1],H[::-1],BASEP)

    elif LAYTYP == LayerType.EQUAL_HEIGHT: # split by equal height intervals
        BASEH = np.linspace(LAYHT, H[-1], NLAY+1)[:-1]
        BASEP = interp(H,P,BASEH)

    elif LAYTYP == LayerType.EQUAL_PATH_LENGTH: # split by equal line-of-sight path intervals
        assert LAYANG<=90 and LAYANG>=0,\
            'Zennith angle should be in [0,90]'
        sin = np.sin(LAYANG*np.pi/180)      # sin(zenith angle)
        cos = np.cos(LAYANG*np.pi/180)      # cos(zenith angle)
        z0 = RADIUS + LAYHT                 # distance from centre to lowest layer's base
        zmax = RADIUS+H[-1]                 # maximum height
        SMAX = np.sqrt(zmax**2-(z0*sin)**2)-z0*cos # total path length
        BASES = np.linspace(0, SMAX, NLAY+1)[:-1]
        BASEH = np.sqrt(BASES**2+z0**2+2*BASES*z0*cos) - RADIUS
        logBASEP = interp(H,np.log(P),BASEH)
        BASEP = np.exp(logBASEP)

    elif LAYTYP == LayerType.BASE_PRESSURE: # split by specifying input base pressures
        if np.isnan(P_base) is True:
            raise ValueError('error in layer_split() :: need input layer base pressures')
        assert  (P_base[-1] >= P[-1]) and (P_base[0] <= P[0]), \
            'Input layer base pressures out of range of atmosphere profile'
        BASEP = P_base
        NLAY = len(BASEP)
        BASEH = interp(P[::-1],H[::-1],BASEP)

    elif LAYTYP == LayerType.BASE_HEIGHT: # split by specifying input base heights
        BASEH = H_base
        logBASEP = interp(H,np.log(P),BASEH)
        BASEP = np.exp(logBASEP)

    else:
        raise('Layering scheme not defined')

    return BASEH, BASEP

#########################################################################################

def read_hlay():

    """
        FUNCTION NAME : read_hlay()
        
        DESCRIPTION : Read the height.lay file used to set the altitude of the base of the layers
                      in a Nemesis run
        
        INPUTS : None
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            nlay :: Number of layers in atmospheric model

        
        CALLING SEQUENCE:
        
            nlay,hbase = read_hlay()
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """

    with open('height.lay','r') as f:

        _ = f.readline().split() # header

        s = f.readline().split()
        nlay = int(s[0])
        hbase = np.zeros(nlay)
        for i in range(nlay):
            s = f.readline().split()
            hbase[i] = float(s[0])

    return nlay,hbase

#########################################################################################


def read_play():

    """
        FUNCTION NAME : read_play()
        
        DESCRIPTION : Read the height.lay file used to set the altitude of the base of the layers
                      in a Nemesis run
        
        INPUTS : None
        
        OPTIONAL INPUTS: none
        
        OUTPUTS :
        
            nlay :: Number of layers in atmospheric model

        
        CALLING SEQUENCE:
        
            nlay,hbase = read_play()
        
        MODIFICATION HISTORY : Juan Alday (29/04/2019)
        
    """

    with open('pressure.lay','r') as f:

        _ = f.readline().split() # header

        s = f.readline().split()
        nlay = int(s[0])
        pbase = np.zeros(nlay)
        for i in range(nlay):
            s = f.readline().split()
            pbase[i] = float(s[0])

    return nlay,pbase

#########################################################################################


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# AtmCalc_0.py - Object to calculate the atmospheric paths.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability
import numpy as np

from archnemesis.enums import ZenithAngleOrigin, PathObserverPointing, PathCalc

import logging
_lgr = logging.getLogger(__name__)

"""
Object to calculate the atmospheric paths
"""

class AtmCalc_0:
    def __init__(
            self,
            Layer,
            path_observer_pointing=PathObserverPointing.DISK,
            #LIMB=False,
            #NADIR=False,
            BOTLAY=0,
            ANGLE=0.0,
            EMISS_ANG=0.0,
            SOL_ANG=0.0,
            AZI_ANG=0.0,
            IPZEN = ZenithAngleOrigin.BOTTOM,
    
            path_calc = PathCalc.PLANCK_FUNCTION_AT_BIN_CENTRE,
            
            #WF=False,
            #NETFLUX=False,
            #OUTFLUX=False,
            #BOTFLUX=False,
            #UPFLUX=False,
            #CG=False,
            #THERM=False,
            #HEMISPHERE=False,
            #NEARLIMB=False,
            #SINGLE=False,
            #SPHSINGLE=False,
            #SCATTER=False,
            #BROAD=False,
            #ABSORB=False,
            #BINBB=True
        ):
        """
            After splitting the atmosphere in different layers and storing them in the Layer class,
            the atmospheric paths are calculated.
            Inputs
            ------
            @param Layer: class
                Python class including all required information about the different atmospheric layers
            @param LIMB: log
                Flag indicating whether it is a limb path. If True, then the attribute BOTLAY must be
                filled with the bottom layer of the path
            @param NADIR: log
                Flag indicating whether it is a nadir path. If True, then the attribute BOTLAY must be
                filled with the bottom layer to use (typically 0) and VIEW_ANG must be filled with the angle
                from the nadir 
            @param BOTLAY: real
                Bottom layer to use in the calculation of the path            
            @param ANGLE: real
                Observing angle from nadir (deg). Note that more than 90deg is looking upwards
            @param EMISS_ANG: real
                Observing angle from nadir (deg). Note that more than 90deg is looking upwards
            @param SOL_ANG: real
                Solar zenith angle (deg). Note that 0 is at zenith ang >90 is below the horizon
            @param AZI_ANG: real
                Azimuth angle (deg). Note that 0 is forward scattering
            @param IPZEN: int
                Flag defining where the zenith angle is defined. 
                0 = at bottom of bottom layer. 
                1 = at the 0km altitude level.
                2 = at the very top of the atmosphere.
            @param WF: log
                Flags indicating the type of calculation to be performed: Weighting function
            @param NETFLUX: log
                Flags indicating the type of calculation to be performed: Net flux calculation
            @param UPFLUX: log
                Flags indicating the type of calculation to be performed: Internal upward flux calculation
            @param OUTFLUX: log
                Flags indicating the type of calculation to be performed: Upward flux at top of topmost layer
            @param BOTFLUX: log
                Flags indicating the type of calculation to be performed: Downward flux at bottom of lowest layer
            @param CG: log
                Flags indicating the type of calculation to be performed: Curtis Godson
            @param THERM: log
                Flags indicating the type of calculation to be performed: Thermal emission
            @param HEMISPHERE: log
                Flags indicating the type of calculation to be performed: Integrate emission into hemisphere
            @param SCATTER: log
                Flags indicating the type of calculation to be performed: Full scattering calculation
            @param NEARLIMB: log
                Flags indicating the type of calculation to be performed: Near-limb scattering calculation
            @param SINGLE: log
                Flags indicating the type of calculation to be performed: Single scattering calculation (plane parallel)
            @param SPHSINGLE: log
                Flags indicating the type of calculation to be performed: Single scattering calculation (spherical atm.)
            @param ABSORB: log
                Flags indicating the type of calculation to be performed: calculate absorption not transmission
            @param BINBB: log
                Flags indicating the type of calculation to be performed: use planck function at bin centre in genlbl
            @param BROAD: log
                Flags indicating the type of calculation to be performed: calculate emission outside of genlbl
            
            Attributes
            -----------
            @attribute SURFACE: log
                Flag indicating whether the observer is position in space (False) and looks downwards or
                in the surface and looks upwards (True)
            @attribute NPATH: int
                Number of atmospheric paths required to perform the atmospheric calculation
            @attribute NLAYIN: 1D array
                For each path, number of layers involved in the calculation
            @attribute LAYINC: 2D array
                For each path, layers involved in the calculation
            @attribute EMTEMP: 1D array
                For each path, emission temperature of the layers involved in the calculation
            @attribute SCALE: 1D array
                For each path, scaling factor to calculate line-of-sight density in each layer with respect
                to the vertical line-of-sight density
            @attribute IMOD: 1D array
                For each path, calculation type       

            Methods
            -------

        """
        _lgr.debug('Sent to AtmCalc_0', stacklevel=2)
        _lgr.debug(f'AtmCalc_0 :: {path_observer_pointing=}, {BOTLAY=}, {ANGLE=}, {EMISS_ANG=}, {SOL_ANG=}, {AZI_ANG=}, {IPZEN=}, {path_calc=}')
        
        #parameters
        self.path_observer_pointing = path_observer_pointing
        self.path_observer_height = np.nan # height of the path observer above height = 0. Can be 0 or np.inf at the moment
        #self.LIMB = LIMB
        #self.NADIR = NADIR
        self.BOTLAY = BOTLAY
        self.ANGLE = ANGLE
        self.EMISS_ANG = EMISS_ANG
        self.SOL_ANG = SOL_ANG
        self.AZI_ANG = AZI_ANG
        self.IPZEN = IPZEN
        self.path_calc = path_calc
        #self.WF = WF
        #self.NETFLUX = NETFLUX
        #self.UPFLUX = UPFLUX
        #self.OUTFLUX = OUTFLUX
        #self.BOTFLUX = BOTFLUX
        #self.CG = CG
        #self.THERM = THERM
        #self.HEMISPHERE = HEMISPHERE
        #self.SCATTER = SCATTER
        #self.NEARLIMB = NEARLIMB
        #self.SINGLE = SINGLE
        #self.SPHSINGLE = SPHSINGLE
        #self.ABSORB = ABSORB
        #self.BINBB = BINBB
        #self.BROAD = BROAD

        #attributes
        #self.SURFACE = None   #Flag indicating whether the observer is on the surface (looking upwards)
        self.NPATH = None     #Number of paths needed for the atmospheric calculation
        self.NLAYIN = None    #np.array(NPATH) For each path, number of layers involved in the calculation
        self.LAYINC = None    #np.array(NLAYIN,NPATH) For each path, layers involved in the calculation
        self.EMTEMP = None    #np.array(NLAYIN,NPATH) For each path, emission temperature of each of the layers involved
        self.SCALE = None     #np.array(NLAYIN,NPATH) For each path, scaling factor 
                              #to calculate line-of-sight density in each layer
        self.IMOD = None      #np.array(NPATH) Calculation type
        self.ITYPE = None
        self.NINTP = None
        self.ICALD = None
        self.NREALP = None
        self.RCALD = None


        #Checking the geometry of the observation 
        ###########################################

        if self.path_observer_pointing == PathObserverPointing.DISK:
            self.path_observer_height = np.inf
            raise NotImplementedError(f'{PathObserverPointing.DISK} is not implemented yet. Intended to be observer pointing towards the disk (i.e. looking down from above). At the moment that case is covered by {PathObserverPointing.LIMB} with `self.ANGLE` set to greater than 90 degrees')
            
        elif self.path_observer_pointing == PathObserverPointing.LIMB:
            self.path_observer_height = np.inf
            self.ANGLE = 90.
        
        elif self.path_observer_pointing == PathObserverPointing.NADIR:
            if self.EMISS_ANG > 90.:
                self.ANGLE = 180.0 - self.ANGLE
                self.path_observer_height = 0.0
            else:
                self.path_observer_height = np.inf
        
        else:
            raise ValueError(f'error in AtmCalc_0 :: path observer pointing "{self.path_observer_pointing}" not recognised')
        
        
        _lgr.debug(f'AtmCalc_0 :: {self.path_observer_pointing=}, {self.BOTLAY=}, {self.ANGLE=}, {self.EMISS_ANG=}, {self.SOL_ANG=}, {self.AZI_ANG=}, {self.IPZEN=}, {self.path_calc=}, {self.path_observer_height=}')
        
        assert self.path_observer_height in (0.0, np.inf), f'error in AtmCalc_0 :: path observer height must be 0 or np.inf. Other values such as {self.path_observer_height} are not implemented yet'

        #Checking incompatible flags and resetting them 
        #################################################

        if (PathCalc.THERMAL_EMISSION in self.path_calc 
                and PathCalc.ABSORBTION in self.path_calc
            ):
            self.path_calc &= ~PathCalc.ABSORBTION
            _lgr.warning(' in AtmCalc_0.py file :: Cannot use absorption for thermal calcs - resetting')

        if (PathCalc.SINGLE_SCATTERING_PLANE_PARALLEL in self.path_calc 
                and PathCalc.MULTIPLE_SCATTERING in self.path_calc
            ):
            self.path_calc &= ~PathCalc.SINGLE_SCATTERING_PLANE_PARALLEL
            _lgr.warning(' in AtmCalc_0.py file :: Cannot use SINGLE and SCATTER - resetting')

        if (PathCalc.SINGLE_SCATTERING_SPHERICAL in self.path_calc 
                and PathCalc.MULTIPLE_SCATTERING in self.path_calc
            ):
            self.path_calc &= ~PathCalc.SINGLE_SCATTERING_SPHERICAL
            _lgr.warning(' in AtmCalc_0.py file :: Cannot use SPHSINGLE and SCATTER - resetting')

        if (PathCalc.PLANCK_FUNCTION_AT_BIN_CENTRE in self.path_calc 
                and PathCalc.BROADENING in self.path_calc
            ):
            self.path_calc &= ~PathCalc.BROADENING
            _lgr.warning(' in AtmCalc_0.py file :: Cannot use BINBB and BROAD - resetting')

        if PathCalc.THERMAL_EMISSION in self.path_calc:
            if PathCalc.BROADENING in self.path_calc:
                self.path_calc &= ~PathCalc.BROADENING
                _lgr.warning(' in AtmCalc_0.py file :: THERM requires BROAD disabled - resetting')
            
            if PathCalc.PLANCK_FUNCTION_AT_BIN_CENTRE in self.path_calc:
                self.path_calc &= ~PathCalc.PLANCK_FUNCTION_AT_BIN_CENTRE
                _lgr.warning(' in AtmCalc_0.py file :: THERM requires BINBB disabled - resetting')
            
        if (((PathCalc.SINGLE_SCATTERING_PLANE_PARALLEL | PathCalc.SINGLE_SCATTERING_SPHERICAL | PathCalc.MULTIPLE_SCATTERING) & self.path_calc) 
                and PathCalc.THERMAL_EMISSION in self.path_calc
            ):
            self.path_calc &= ~PathCalc.THERMAL_EMISSION
            _lgr.info('THERM not required. Scattering includes emission')

        if (PathCalc.HEMISPHERE in self.path_calc
                and PathCalc.THERMAL_EMISSION not in self.path_calc
            ):
            self.path_calc |= PathCalc.THERMAL_EMISSION
            _lgr.warning(' in AtmCalc_0.py file :: HEMISPHERE assumes THERM - resetting')

        if (((PathCalc.SINGLE_SCATTERING_PLANE_PARALLEL | PathCalc.SINGLE_SCATTERING_SPHERICAL | PathCalc.MULTIPLE_SCATTERING) & self.path_calc) 
                and PathCalc.CURTIS_GODSON in self.path_calc
            ):
            self.path_calc &= ~PathCalc.CURTIS_GODSON
            _lgr.warning(' in AtmCalc_0.py file :: Cannot use CG and SCATTER - resetting')

        if (PathCalc.SINGLE_SCATTERING_PLANE_PARALLEL | PathCalc.SINGLE_SCATTERING_SPHERICAL | PathCalc.MULTIPLE_SCATTERING) & self.path_calc:
            if self.path_observer_pointing == PathObserverPointing.LIMB:
                if PathCalc.SINGLE_SCATTERING_PLANE_PARALLEL in self.path_calc:
                    raise ValueError('error in AtmCalc_0.py file :: SINGLE and LIMB not catered for')
                if PathCalc.SINGLE_SCATTERING_SPHERICAL in self.path_calc:
                    raise ValueError('error in AtmCalc_0.py file :: SPHSINGLE and LIMB not catered for')  
            else:
                if self.ANGLE != 0.0:
                    _lgr.warning(' in AtmCalc_0.py file :: ANGLE must be 0.0 for scattering calculations - resetting')
                    self.ANGLE = 0.0

        if PathCalc.HEMISPHERE in self.path_calc:
            if self.path_observer_pointing == PathObserverPointing.LIMB:
                raise ValueError('error in AtmCalc_0.py file :: cannot do HEMISPHERE and LIMB')  
            else:
                if self.ANGLE != 0.0:
                    _lgr.warning(' in AtmCalc_0.py file :: ANGLE must be 0.0 for HEMISPHERE - resetting') 
                    self.ANGLE = 0.0
        
        
        #Translating ANGLE to be defined at bottom of bottom layer in case it
        #has been defined in another way in the AtmCalc_0.py file
        #######################################################################

        if(self.IPZEN==ZenithAngleOrigin.ALTITUDE_ZERO):
            #Compute zenith angle of ray at bottom of bottom layer, assuming it
            #has been defined at the 0km level
            z0 = Layer.RADIUS + Layer.BASEH[self.BOTLAY]
            self.ANGLE = np.arcsin(Layer.RADIUS/z0 * np.sin(self.ANGLE/180.*np.pi)) / np.pi * 180.
        elif(self.IPZEN==ZenithAngleOrigin.TOP):
            #Compute zenith angle of ray at bottom of bottom layer, assuming it
            #has been defined at the top of the atmosphere
            z0 = Layer.RADIUS + Layer.BASEH[Layer.NLAY-1] + Layer.DELH[Layer.NLAY-1]

            #Calculate tangent altitude of ray at lowest point
            HTAN = z0*np.sin(self.ANGLE/180.*np.pi)-Layer.RADIUS

            if HTAN<=Layer.BASEH[self.BOTLAY]:
                #Calculate zenith angle at bottom of lowest layer
                self.ANGLE = np.arcsin(z0/(Layer.RADIUS + Layer.BASEH[self.BOTLAY]) * np.sin(self.ANGLE/180.*np.pi)) / np.pi * 180.
            else:
                #We need to model this ray as a tangent path.
                self.path_observer_pointing = PathObserverPointing.LIMB
                self.ANGLE = 90.

                #Find number of bottom layer. Snap to layer with nearest base height
                #to computed tangent height.
                for ILAY in range(Layer.NLAY):
                    if Layer.BASEH[ILAY]<HTAN:
                        self.BOTLAY = ILAY
                
                if self.BOTLAY<Layer.NLAY-1:
                    F = (HTAN-Layer.BASEH[self.BOTLAY])/(Layer.BASEH[self.BOTLAY+1]-Layer.BASEH[self.BOTLAY])
                    if F>0.5:
                        self.BOTLAY = self.BOTLAY + 1

        Z0 = Layer.RADIUS + Layer.BASEH[self.BOTLAY]
        SIN2A = np.sin(self.ANGLE/180.*np.pi)**2.
        COSA = np.cos(self.ANGLE/180.*np.pi)

        #Calculate which layers to use in the calculation
        #######################################################################

        #Limb path
        if self.path_observer_pointing == PathObserverPointing.LIMB:

            #Calculate the total number of layers to use
            NUSE = int(2*(Layer.NLAY-self.BOTLAY))

            #Locating the layers to be included
            USELAY = np.zeros(NUSE,dtype='int32')
            for IUSE in range(int(NUSE/2)):
                USELAY[IUSE] = Layer.NLAY - 1 - IUSE
                USELAY[int(NUSE/2)+IUSE] = self.BOTLAY + IUSE
                
            #Calculating the emission temperature for those layers
            EMITT = np.zeros(NUSE)  #Emission temperature
            for IUSE in range(NUSE):
                EMITT[IUSE] = Layer.TEMP[USELAY[IUSE]]

        #Nadir path
        if self.path_observer_pointing == PathObserverPointing.NADIR:

            #Calculate the total number of layers to use
            NUSE = Layer.NLAY-self.BOTLAY

            if self.path_observer_height == 0.0: # Observer on the surface looking upwards
                #Locating layers and calculating emission temperature
                USELAY = np.zeros(NUSE,dtype='int32')
                EMITT = np.zeros(NUSE)  #Emission temperature
                for IUSE in range(NUSE):
                    USELAY[IUSE] = IUSE 
                    EMITT[IUSE] = Layer.TEMP[USELAY[IUSE]]

            elif self.path_observer_height == np.inf: # Observer in space looking down

                #Locating layers and calculating emission temperature
                USELAY = np.zeros(NUSE,dtype='int32')
                EMITT = np.zeros(NUSE)  #Emission temperature
                for IUSE in range(NUSE):
                    USELAY[IUSE] = Layer.NLAY - 1 - IUSE 
                    EMITT[IUSE] = Layer.TEMP[USELAY[IUSE]]


        #Computing the scale factor for the path in each layer (with respect to vertical integration)
        #################################################################################################

        SF = np.zeros(NUSE)   #Scaling factor 
        for IUSE in range(NUSE):

            STMP = (Layer.RADIUS + Layer.BASEH[USELAY[IUSE]])**2. - SIN2A * Z0**2.
            #Sometimes, there are rounding errors here that cause the program
            #to try to take the square root of a _very_ small negative number.
            #This quietly fixes that and hopefully doesn't break anything else.

            if STMP<0.0:
                STMP = 0.0
            
            S0 = np.sqrt(STMP) - Z0 * COSA

            if USELAY[IUSE]<Layer.NLAY - 1:
                S1 = np.sqrt( (Layer.RADIUS+Layer.BASEH[USELAY[IUSE]+1])**2. - SIN2A * Z0**2. ) - Z0 * COSA
                SF[IUSE] = (S1-S0)/(Layer.BASEH[USELAY[IUSE]+1]-Layer.BASEH[USELAY[IUSE]])
            if USELAY[IUSE]==Layer.NLAY - 1:
                NPRO = len(Layer.H)
                S1 = np.sqrt( (Layer.RADIUS+Layer.H[NPRO-1])**2. - SIN2A * Z0**2. ) - Z0 * COSA
                SF[IUSE] = (S1-S0)/(Layer.H[NPRO-1]-Layer.BASEH[USELAY[IUSE]])


        #Calculating any Curtis-Godson paths if needed
        #####################################################

        if PathCalc.CURTIS_GODSON in self.path_calc:
            raise NotImplementedError('error in AtmCalc_0.py file :: Curtis-Godson files are not implemented in the path calculation yet')

        #Calculating the calculation type to pass to RADTRANS
        ######################################################

        #Calculating the number of paths required for the calculation
        #For example if calculating a weighting function or if performing a thermal integration outside the main Radtrans routines

        self.NPATH = 1

        if PathCalc.WEIGHTING_FUNCTION in self.path_calc:
            self.NPATH = NUSE

        if (PathCalc.THERMAL_EMISSION in self.path_calc
                and PathCalc.BROADENING in self.path_calc
            ):
            self.NPATH = NUSE

        if PathCalc.UPWARD_FLUX in self.path_calc:
            self.NPATH = NUSE

        if PathCalc.NET_FLUX in self.path_calc:
            raise ValueError('error :: need to properly define the paths (should be 2*NLAYER for upward and downward flux)')
            self.NPATH = NUSE

        if (PathCalc.UPWARD_FLUX in self.path_calc
                and PathCalc.MULTIPLE_SCATTERING not in self.path_calc
            ):
            raise ValueError('error in AtmCalc_0.py file :: cannot do upward flux calculation with scattering turned off')

        if (PathCalc.OUTWARD_FLUX in self.path_calc
                and PathCalc.MULTIPLE_SCATTERING not in self.path_calc
            ):
            raise ValueError('error in AtmCalc_0.py file :: cannot do outward flux calculation with scattering turned off')

        if (PathCalc.DOWNWARD_FLUX in self.path_calc
                and PathCalc.MULTIPLE_SCATTERING not in self.path_calc
            ):
            raise ValueError('error in AtmCalc_0.py file :: cannot do downward flux calculation with scattering turned off')

        NLAYIN = np.zeros(self.NPATH,dtype='int32')
        LAYINC = np.zeros([NUSE,self.NPATH],dtype='int32')
        SCALE = np.zeros([NUSE,self.NPATH])
        EMTEMP = np.zeros([NUSE,self.NPATH])
        
        # [JD] From what I can tell this loop only uses static values, there is
        # no need to loop over the number of paths to create IMOD
        assert len(PathCalc) <= 32, f'error in AtmCalc_0.py file :: enums.PathCalc must have less than 32 flags to fit into a np.int32, it has {len(~PathCalc(0))}. Either increase the size of the integer type holding them, or reduce the number of flags.'
        IMOD = np.full((self.NPATH,), fill_value=self.path_calc, dtype=np.int32)
        
        for j in range(self.NPATH):
            
            if PathCalc.CURTIS_GODSON in self.path_calc:
                raise NotImplementedError('error in AtmCalc_0.py file :: Curtis-Godson files are not implemented in the path calculation yet')
            else:
                NLAYIN[j] = (j+1) + NUSE - self.NPATH 
                # NLAYIN chosen so that if NPATH=1, use layers 1 to NUSE but
                # if NPATH=NUSE then include paths 1 to J. 
                # i.e. 1 to 1, 1 to 2, up to 1 to NUSE

                for i in range(NLAYIN[j]):
                    LAYINC[i,j] = USELAY[i]
                    EMTEMP[i,j] = EMITT[i]
                    SCALE[i,j] = SF[i]



        self.IMOD = IMOD
        self.NLAYIN = NLAYIN
        self.LAYINC = LAYINC
        self.EMTEMP = EMTEMP
        self.SCALE = SCALE

        #Having calculated the atmospheric paths, now outputing the calculation
        #############################################################################

        ITYPE = self.path_calc
        
        NINTP = 3
        ICALD = np.zeros(NINTP,dtype='int32')
        ICALD[0] = 1
        ICALD[1] = self.NPATH
        ICALD[2] = self.BOTLAY
        NREALP = 2
        RCALD = np.zeros(NREALP,dtype='int32')
        RCALD[0] = ANGLE
        HT = 0.0        # Fix!!!!!!!!!!!!!!!
        RCALD[1] = HT

        self.ITYPE = ITYPE
        self.NINTP = NINTP
        self.ICALD = ICALD
        self.NREALP = NREALP
        self.RCALD = RCALD